#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2017 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import os
import signal
import six
from io import TextIOWrapper

import pyuv

from . import fibers, compat, futures
from .hub import get_hub, switchpoint
from .sync import Event
from .errors import Timeout
from .stream import StreamProtocol
from .endpoints import create_connection, Endpoint
from .callbacks import add_callback, remove_callback, run_callbacks

__all__ = ['Process', 'PIPE', 'DEVNULL']

PIPE = -1
DEVNULL = -3

CREATE_PIPE = pyuv.UV_CREATE_PIPE | pyuv.UV_READABLE_PIPE | pyuv.UV_WRITABLE_PIPE


class Process(Endpoint):
    """A child process.

    This class allows you to start up a child process, communicate with it and
    control it. The API is modeled after the :class:`subprocess.Popen` class.
    """

    def __init__(self, encoding=None, textio_args={}, timeout=None):
        """
        The *encoding* argument specifies the encoding to use for output of the
        child. If it is not specified, then reading from the child will produce
        ``bytes`` objects.

        The *textio_args* argument can be used to pass keyword arguments to the
        :class:`io.TextIOWrapper` instances that are used to wrap the raw
        standard input and outputs in case *encoding* is provided. It can be
        used e.g. to change buffering and enable universal newlines.
        """
        super(Process, self).__init__(StreamProtocol, timeout)
        self._encoding = encoding
        self._textio_args = textio_args.copy()
        # Unless specified otherwise, a text stream should be write-through.
        self._emulate_write_through = False
        if encoding and 'line_buffering' not in textio_args \
                    and 'write_through' not in textio_args:
            if six.PY2:
                self._emulate_write_through = True
            else:
                textio_args['write_through'] = True
        self._process = None
        self._child_exited = Event()
        self._closed = Event()
        self._stdin = self._stdout = self._stderr = None
        self._exit_status = None
        self._term_signal = None
        self._callbacks = None

    @property
    def stdin(self):
        """The child's standard input, or None."""
        if self._stdin:
            return self._stdin[0]

    @property
    def stdout(self):
        """The child's standard output, or None."""
        if self._stdout:
            return self._stdout[0]

    @property
    def stderr(self):
        """The child's standard error, or None."""
        if self._stderr:
            return self._stderr[0]

    @property
    def returncode(self):
        """The child's exit status, or None if it has not exited yet.

        On Unix, if the child was terminated by a signal, return  ``-SIGNUM``.
        """
        if self._term_signal:
            return -self._term_signal
        return self._exit_status

    @property
    def pid(self):
        """The child's process ID, or None if there is no child."""
        if self._process is None:
            return
        return self._process.pid

    def _create_stdio(self, loop, spec, childfd):
        # Create a pyuv.StdIO container
        if spec is None:
            stdio = pyuv.StdIO(fd=childfd, flags=pyuv.UV_INHERIT_FD)
        elif spec == PIPE:
            stdio = pyuv.StdIO(stream=pyuv.Pipe(loop), flags=CREATE_PIPE)
        elif spec == DEVNULL:
            stdio = pyuv.StdIO(flags=pyuv.UV_IGNORE)
        elif isinstance(spec, pyuv.Stream):
            stdio = pyuv.StdIO(stream=spec, flags=CREATE_PIPE)
        elif isinstance(spec, int) and spec >= 0:
            stdio = pyuv.StdIO(fd=spec, flags=pyuv.UV_INHERIT_FD)
        elif hasattr(spec, 'fileno'):
            stdio = pyuv.StdIO(fd=spec.fileno(), flags=CREATE_PIPE)
        else:
            raise TypeError('stdio[{}]: spec must be PIPE, an fd, a Stream, or a file-like object'
                            ' (got {!r})'.format(childfd, type(spec).__name__))
        return stdio

    def _get_child_handles(self, loop, stdin, stdout, stderr, extra_handles):
        # Return a list of StdIO containers that are passed to our child
        handles = []
        handles.append(self._create_stdio(loop, stdin, 0))
        handles.append(self._create_stdio(loop, stdout, 1))
        handles.append(self._create_stdio(loop, stderr, 2))
        for ix, spec in enumerate(extra_handles or []):
            if spec is None or spec == PIPE:
                raise TypeError('extra_handles[{}]: cannot be None or PIPE'.format(ix))
            handles.append(self._create_stdio(loop, spec, ix+3))
        return handles

    def _connect_child_handle(self, stdio):
        # Connect a StdIO container to a StreamProtocol. Return the (possibly
        # textio wrapped) Stream
        transport, protocol = create_connection(StreamProtocol, stdio.stream,
                                                timeout=self._timeout)
        if self._encoding:
            stream = TextIOWrapper(protocol.stream, self._encoding, **self._textio_args)
            if self._emulate_write_through:
                compat.make_textiowrapper_writethrough(stream)
        else:
            stream = protocol.stream
        return stream, transport, protocol

    def spawn(self, args, executable=None, stdin=None, stdout=None, stderr=None,
              shell=False, cwd=None, env=None, flags=0, extra_handles=None):
        """Spawn a new child process.

        The executable to spawn and its arguments are determined by *args*,
        *executable* and *shell*.

        When *shell* is set to ``False`` (the default), *args* is normally a
        sequence and it contains both the program to execute (at index 0), and
        its arguments.

        When *shell* is set to ``True``, then *args* is normally a string and
        it indicates the command to execute through the shell.

        The *executable* argument can be used to override the executable to
        execute. If *shell* is ``False``, it overrides ``args[0]``. This is
        sometimes used on Unix to implement "fat" executables that behave
        differently based on argv[0]. If *shell* is ``True``, it overrides the
        shell to use. The default shell is ``'/bin/sh'`` on Unix, and the value
        of $COMSPEC (or ``'cmd.exe'`` if it is unset) on Windows.

        The *stdin*, *stdout* and *stderr* arguments specify how to handle
        standard input, output, and error, respectively. If set to None, then
        the child will inherit our respective stdio handle. If set to the
        special constant ``PIPE`` then a pipe is created. The pipe will be
        connected to a :class:`gruvi.StreamProtocol` which you can use to read
        or write from it. The stream protocol instance is available under
        either :attr:`stdin`, :attr:`stdout` or :attr:`stderr`. All 3 stdio
        arguments can also be a file descriptor, a file-like object, or a pyuv
        ``Stream`` instance.

        The *extra_handles* specifies any extra handles to pass to the client.
        It must be a sequence where each element is either a file descriptor, a
        file-like objects, or a ``pyuv.Stream`` instance. The position in the
        sequence determines the file descriptor in the client. The first
        position corresponds to FD 3, the second to 4, etc. This places these
        file descriptors directly after the stdio handles.

        The *cwd* argument specifies the directory to change to before
        executing the child. If not provided, the current directory is used.

        The *env* argument specifies the environment to use when executing the
        child. If provided, it must be a dictionary. By default, the current
        environment is used.

        The *flags* argument can be used to specify optional libuv
        ``uv_process_flags``. The only relevant flags are
        ``pyuv.UV_PROCESS_DETACHED`` and ``pyuv.UV_PROCESS_WINDOWS_HIDE``. Both
        are Windows specific and are silently ignored on Unix.
        """
        if self._process:
            raise RuntimeError('child process already spawned')
        self._child_exited.clear()
        self._closed.clear()
        self._exit_status = None
        self._term_signal = None
        hub = get_hub()
        if isinstance(args, str):
            args = [args]
            flags |= pyuv.UV_PROCESS_WINDOWS_VERBATIM_ARGUMENTS
        else:
            args = list(args)
        if shell:
            if hasattr(os, 'fork'):
                # Unix
                if executable is None:
                    executable = '/bin/sh'
                args = [executable, '-c'] + args
            else:
                # Windows
                if executable is None:
                    executable = os.environ.get('COMSPEC', 'cmd.exe')
                args = [executable, '/c'] + args
        if executable is None:
            executable = args[0]
        kwargs = {}
        if env is not None:
            kwargs['env'] = env
        if cwd is not None:
            kwargs['cwd'] = cwd
        kwargs['flags'] = flags
        handles = self._get_child_handles(hub.loop, stdin, stdout, stderr, extra_handles)
        kwargs['stdio'] = handles
        process = pyuv.Process.spawn(hub.loop, args, executable,
                                     exit_callback=self._on_child_exit, **kwargs)
        # Create stdin/stdout/stderr transports/protocols.
        if handles[0].stream:
            self._stdin = self._connect_child_handle(handles[0])
        if handles[1].stream:
            self._stdout = self._connect_child_handle(handles[1])
        if handles[2].stream:
            self._stderr = self._connect_child_handle(handles[2])
        self._process = process

    def _on_child_exit(self, handle, exit_status, term_signal):
        # Callback used as the exit_callback with pyuv.Process
        self._exit_status = exit_status
        self._term_signal = term_signal
        self._child_exited.set()
        self.child_exited(exit_status, term_signal)
        run_callbacks(self)

    def child_exited(self, exit_status, term_signal):
        """Callback that is called when the child has exited."""

    def _on_close_complete(self, handle):
        # Callback used as the callback with Process.close()
        self._closed.set()

    @switchpoint
    def close(self):
        """Close the process and frees its associated resources.

        This method waits for the resources to be freed by the event loop.
        """
        if self._process is None:
            return
        waitfor = []
        if not self._process.closed:
            self._process.close(self._on_close_complete)
            waitfor.append(self._closed)
        # For each of stdin/stdout/stderr, close the transport. This schedules
        # an on-close callback that will close the protocol, which we wait for.
        if self._stdin:
            self._stdin[1].close()
            waitfor.append(self._stdin[1]._closed)
        if self._stdout:
            self._stdout[1].close()
            waitfor.append(self._stdout[1]._closed)
        if self._stderr:
            self._stderr[1].close()
            waitfor.append(self._stderr[1]._closed)
        futures.wait(waitfor)
        self._process = None
        self._stdin = self._stdout = self._stderr = None

    def send_signal(self, signum):
        """Send the signal *signum* to the child.

        On Windows, SIGTERM, SIGKILL and SIGINT are emulated using
        TerminateProcess(). This will cause the child to exit unconditionally
        with status 1. No other signals can be sent on Windows.
        """
        if self._process is None:
            raise RuntimeError('no child process')
        self._process.kill(signum)

    def terminate(self):
        """Terminate the child process.

        It is not an error to call this method when the child has already exited.
        """
        try:
            self.send_signal(signal.SIGTERM)
        except pyuv.error.ProcessError as e:
            if e.args[0] != pyuv.errno.UV_ESRCH:
                raise

    @switchpoint
    def wait(self, timeout=-1):
        """Wait for the child to exit.

        Wait for at most *timeout* seconds, or indefinitely if *timeout* is
        None. Return the value of the :attr:`returncode` attribute.
        """
        if self._process is None:
            raise RuntimeError('no child process')
        if timeout == -1:
            timeout = self._timeout
        if not self._child_exited.wait(timeout):
            raise Timeout('timeout waiting for child to exit')
        return self.returncode

    @switchpoint
    def communicate(self, input=None, timeout=-1):
        """Communicate with the child and return its output.

        If *input* is provided, it is sent to the client. Concurrent with
        sending the input, the child's standard output and standard error are
        read, until the child exits.

        The return value is a tuple ``(stdout_data, stderr_data)`` containing
        the data read from standard output and standard error.
        """
        if self._process is None:
            raise RuntimeError('no child process')
        if timeout == -1:
            timeout = self._timeout
        output = [[], []]
        def writer(stream, data):
            offset = 0
            while offset < len(data):
                buf = data[offset:offset+4096]
                stream.write(buf)
                offset += len(buf)
            stream.close()
        def reader(stream, data):
            while True:
                if self._encoding:
                    buf = stream.read(4096)
                else:
                    buf = stream.read1()
                if not buf:
                    break
                data.append(buf)
        if self.stdin:
            fibers.spawn(writer, self.stdin, input or b'')
        if self.stdout:
            fibers.spawn(reader, self.stdout, output[0])
        if self.stderr:
            fibers.spawn(reader, self.stderr, output[1])
        self.wait(timeout)
        empty = '' if self._encoding else b''
        stdout_data = empty.join(output[0])
        stderr_data = empty.join(output[1])
        return (stdout_data, stderr_data)

    # Support wait()

    def add_done_callback(self, callback, *args):
        return add_callback(self._child_exited, callback, args)

    def remove_done_callback(self, handle):
        remove_callback(self._child_exited, handle)



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2017 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import pyuv
import ssl

from . import compat
from .transports import Transport, TransportError
from .sync import Event

from .sslcompat import SSLContext, MemoryBIO, get_reason, wrap_bio

__all__ = ['SslTransport', 'create_ssl_context']


class SslPipe(object):
    """An SSL "Pipe".

    An SSL pipe allows you to communicate with an SSL/TLS protocol instance
    through memory buffers. It can be used to implement a security layer for an
    existing connection where you don't have access to the connection's file
    descriptor, or for some reason you don't want to use it.

    An SSL pipe can be in "wrapped" and "unwrapped" mode. In unwrapped mode,
    data is passed through untransformed. In wrapped mode, application level
    data is encrypted to SSL record level data and vice versa. The SSL record
    level is the lowest level in the SSL protocol suite and is what travels
    as-is over the wire.

    An SslPipe initially is in "unwrapped" mode. To start SSL, call
    :meth:`do_handshake`. To shutdown SSL again, call :meth:`unwrap`.
    """

    bufsize = 65536

    # This previously used a socketpair to communicate with the SSL protocol
    # instance but since October 2014 we're using a Memory BIO! This is
    # cleaner, and more reliable on Windows. See for example issue #12 for more
    # details.

    S_UNWRAPPED, S_DO_HANDSHAKE, S_WRAPPED, S_SHUTDOWN = range(4)

    def __init__(self, context, server_side, server_hostname=None):
        """
        The *context* argument specifies the :class:`ssl.SSLContext` to use.
        It is recommended to use :func:`~gruvi.ssl.create_ssl_context` so that
        it will work on all supported Python versions.

        The *server_side* argument indicates whether this is a server side or
        client side transport.

        The optional *server_hostname* argument can be used to specify the
        hostname you are connecting to. You may only specify this parameter if
        the _ssl module supports Server Name Indication (SNI).
        """
        self._context = context
        self._server_side = server_side
        self._server_hostname = server_hostname
        self._state = self.S_UNWRAPPED
        self._bios = (MemoryBIO(), MemoryBIO())
        self._sslobj = None
        self._need_ssldata = False

    @property
    def context(self):
        """The SSL context passed to the constructor."""
        return self._context

    @property
    def ssl(self):
        """The internal :class:`ssl.SSLObject` instance."""
        return self._sslobj

    @property
    def need_ssldata(self):
        """Whether more record level data is needed to complete a handshake
        that is currently in progress."""
        return self._need_ssldata

    @property
    def wrapped(self):
        """Whether a security layer is currently in effect."""
        return self._state == self.S_WRAPPED

    def do_handshake(self, callback=None):
        """Start the SSL handshake. Return a list of ssldata.

        The optional *callback* argument can be used to install a callback that
        will be called when the handshake is complete. The callback will be
        called without arguments.
        """
        if self._state != self.S_UNWRAPPED:
            raise RuntimeError('handshake in progress or completed')
        wrapargs = (self._bios[0], self._bios[1], self._server_side, self._server_hostname)
        self._sslobj = wrap_bio(self._context, *wrapargs)
        self._state = self.S_DO_HANDSHAKE
        self._on_handshake_complete = callback
        ssldata, appdata = self.feed_ssldata(b'')
        assert len(appdata) == 0
        return ssldata

    def shutdown(self, callback=None):
        """Start the SSL shutdown sequence. Return a list of ssldata.

        The optional *callback* argument can be used to install a callback that
        will be called when the shutdown is complete. The callback will be
        called without arguments.
        """
        if self._state == self.S_UNWRAPPED:
            raise RuntimeError('no security layer present')
        self._state = self.S_SHUTDOWN
        self._on_handshake_complete = callback
        ssldata, appdata = self.feed_ssldata(b'')
        assert appdata == [] or appdata == [b'']
        return ssldata

    def feed_eof(self):
        """Send a potentially "ragged" EOF.

        This method will raise an SSL_ERROR_EOF exception if the EOF is
        unexpected.
        """
        self._bios[0].write_eof()
        ssldata, appdata = self.feed_ssldata(b'')
        assert appdata == [] or appdata == [b'']

    def feed_ssldata(self, data):
        """Feed SSL record level data into the pipe.

        The data must be a bytes instance. It is OK to send an empty bytes
        instance. This can be used to get ssldata for a handshake initiated by
        this endpoint.

        Return a (ssldata, appdata) tuple. The ssldata element is a list of
        buffers containing SSL data that needs to be sent to the remote SSL.

        The appdata element is a list of buffers containing plaintext data that
        needs to be forwarded to the application. The appdata list may contain
        an empty buffer indicating an SSL "close_notify" alert. This alert must
        be acknowledged by calling :meth:`shutdown`.
        """
        if self._state == self.S_UNWRAPPED:
            # If unwrapped, pass plaintext data straight through.
            return ([], [data] if data else [])
        ssldata = []; appdata = []
        self._need_ssldata = False
        if data:
            self._bios[0].write(data)
        try:
            if self._state == self.S_DO_HANDSHAKE:
                # Call do_handshake() until it doesn't raise anymore.
                self._sslobj.do_handshake()
                self._state = self.S_WRAPPED
                if self._on_handshake_complete:
                    self._on_handshake_complete()
            if self._state == self.S_WRAPPED:
                # Main state: read data from SSL until close_notify
                while True:
                    chunk = self._sslobj.read(self.bufsize)
                    appdata.append(chunk)
                    if not chunk:  # close_notify
                        break
            if self._state == self.S_SHUTDOWN:
                # Call shutdown() until it doesn't raise anymore.
                self._sslobj.unwrap()
                self._sslobj = None
                self._state = self.S_UNWRAPPED
                if self._on_handshake_complete:
                    self._on_handshake_complete()
            if self._state == self.S_UNWRAPPED:
                # Drain possible plaintext data after close_notify.
                appdata.append(self._bios[0].read())
        except ssl.SSLError as e:
            if e.errno not in (ssl.SSL_ERROR_WANT_READ, ssl.SSL_ERROR_WANT_WRITE,
                               ssl.SSL_ERROR_SYSCALL):
                raise
            self._need_ssldata = e.errno == ssl.SSL_ERROR_WANT_READ
        # Check for record level data that needs to be sent back.
        # Happens for the initial handshake and renegotiations.
        if self._bios[1].pending:
            ssldata.append(self._bios[1].read())
        return (ssldata, appdata)

    def feed_appdata(self, data, offset=0):
        """Feed plaintext data into the pipe.

        Return an (ssldata, offset) tuple. The ssldata element is a list of
        buffers containing record level data that needs to be sent to the
        remote SSL instance. The offset is the number of plaintext bytes that
        were processed, which may be less than the length of data.

        NOTE: In case of short writes, this call MUST be retried with the SAME
        buffer passed into the *data* argument (i.e. the ``id()`` must be the
        same). This is an OpenSSL requirement. A further particularity is that
        a short write will always have offset == 0, because the _ssl module
        does not enable partial writes. And even though the offset is zero,
        there will still be encrypted data in ssldata.
        """
        if self._state == self.S_UNWRAPPED:
            # pass through data in unwrapped mode
            return ([data[offset:]] if offset < len(data) else [], len(data))
        ssldata = []
        view = memoryview(data)
        while True:
            self._need_ssldata = False
            try:
                if offset < len(view):
                    offset += self._sslobj.write(view[offset:])
            except ssl.SSLError as e:
                # It is not allowed to call write() after unwrap() until the
                # close_notify is acknowledged. We return the condition to the
                # caller as a short write.
                if get_reason(e) == 'PROTOCOL_IS_SHUTDOWN':
                    e.errno = ssl.SSL_ERROR_WANT_READ
                if e.errno not in (ssl.SSL_ERROR_WANT_READ, ssl.SSL_ERROR_WANT_WRITE,
                                   ssl.SSL_ERROR_SYSCALL):
                    raise
                self._need_ssldata = e.errno == ssl.SSL_ERROR_WANT_READ
            # See if there's any record level data back for us.
            if self._bios[1].pending:
                ssldata.append(self._bios[1].read())
            if offset == len(view) or self._need_ssldata:
                break
        return (ssldata, offset)


class SslTransport(Transport):
    """An SSL/TLS transport."""

    def __init__(self, handle, context, server_side, server_hostname=None,
                 do_handshake_on_connect=True, close_on_unwrap=True):
        """
        The *handle* argument is the pyuv handle on top of which to layer the
        SSL transport. It must be a ``pyuv.Stream`` instance, so either a
        :class:`pyuv.TCP`, :class:`pyuv.Pipe` or a :class:`pyuv.TTY`.

        SSL transports are always read-write, so the handle provided needs
        to support that.

        The *context* argument specifies the :class:`ssl.SSLContext` to use.
        You can use :func:`create_ssl_context` to create a new context which
        also works on Python 2.x where :class:`ssl.SSLContext` does not exist.

        The *server_side* argument specifies whether this is a server side or a
        client side transport.

        The optional *server_hostname* argument can be used to specify the
        hostname you are connecting to. You may only specify this parameter if
        your Python version supports SNI.

        The optional *do_handshake_on_connect* argument specifies whether to to
        start the SSL handshake immediately. If False, then the connection will
        be unencrypted until :meth:`do_handshake` is called. The default is to
        start the handshake immediately.

        The optional *close_on_unwrap* argument specifies whether you want the
        ability to continue using the connection after you call :meth:`unwrap`
        of when an SSL "close_notify" is received from the remote peer. The
        default is to close the connection.
        """
        super(SslTransport, self).__init__(handle)
        self._sslpipe = SslPipe(context, server_side, server_hostname)
        self._do_handshake_on_connect = do_handshake_on_connect
        self._close_on_unwrap = close_on_unwrap
        self._write_backlog = []
        self._ssl_active = Event()

    def start(self, protocol):
        # Bind to *protocol* and start calling callbacks on it.
        events = super(SslTransport, self).start(protocol)
        if self._do_handshake_on_connect:
            self.do_handshake()
            if events is None:
                events = []
            events.append(self._ssl_active)
        return events

    def get_extra_info(self, name, default=None):
        """Return transport specific data.

        The following fields are available, in addition to the information
        exposed by :meth:`Transport.get_extra_info`.

        ======================  ===============================================
        Name                    Description
        ======================  ===============================================
        ``'ssl'``               The internal ``ssl.SSLObject`` instance used by
                                this transport.
        ``'sslctx'``            The ``ssl.SSLContext`` instance used to create
                                the SSL object.
        ======================  ===============================================
        """
        if name == 'ssl':
            return self._sslpipe.ssl
        elif name == 'sslctx':
            return self._sslpipe.context
        else:
            return super(SslTransport, self).get_extra_info(name, default)

    def write(self, data):
        # Write *data* to the transport.
        if not isinstance(data, (bytes, bytearray, memoryview)):
            raise TypeError("data: expecting a bytes-like instance, got {!r}"
                                .format(type(data).__name__))
        if self._error:
            raise compat.saved_exc(self._error)
        elif self._closing or self._handle.closed:
            raise TransportError('transport is closing/closed')
        elif len(data) == 0:
            return
        self._write_backlog.append([data, 0])
        self._process_write_backlog()

    def _process_write_backlog(self):
        # Try to make progress on the write backlog.
        try:
            for i in range(len(self._write_backlog)):
                data, offset = self._write_backlog[0]
                if data:
                    ssldata, offset = self._sslpipe.feed_appdata(data, offset)
                elif offset:
                    ssldata, offset = self._sslpipe.do_handshake(self._ssl_active.set), 1
                else:
                    ssldata, offset = self._sslpipe.shutdown(self._ssl_active.clear), 1
                # Temporarily set _closing to False to prevent Transport.write()
                # from raising an error when we are doing a close_notify.
                saved, self._closing = self._closing, False
                # Write the ssl data that came out of the SSL pipe to the handle.
                # Note that flow control is done at the record level data.
                for chunk in ssldata:
                    super(SslTransport, self).write(chunk)
                self._closing = saved
                if offset < len(data):
                    self._write_backlog[0][1] = offset
                    # A short write means that a write is blocked on a read
                    # We need to enable reading if it is not enabled!!
                    # At the same time stop the protocol from writing data to
                    # make the situation not worse.
                    assert self._sslpipe.need_ssldata
                    if not self._reading:
                        self.resume_reading()
                    if self._writing:
                        self._protocol.pause_writing()
                        self._writing = False
                    break
                # An entire chunk from the backlog was processed. We can
                # delete it and reduce the outstanding buffer size.
                del self._write_backlog[0]
        except ssl.SSLError as e:
            self._log.warning('SSL error {} (reason {})', e.errno, e.reason, exc_info=True)
            self._error = e
            self.abort()

    def _on_read_complete(self, handle, data, error):
        # Callback used with handle.start_read().
        assert handle is self._handle
        try:
            if self._error:
                self._log.warning('ignore read status {} after close', error)
            elif error == pyuv.errno.UV_EOF:
                self._sslpipe.feed_eof()  # Raises SSL_ERROR_EOF if unexpected
                if self._protocol.eof_received() and not self._close_on_unwrap:
                    self._log.debug('EOF received, protocol wants to continue')
                elif not self._closing:
                    self._log.debug('EOF received, closing transport')
                    super(SslTransport, self).close()
            elif error:
                self._log.warning('pyuv error {} in read callback', error)
                self._error = TransportError.from_errno(error)
                self.abort()
            else:
                ssldata, appdata = self._sslpipe.feed_ssldata(data)
                for chunk in ssldata:
                    super(SslTransport, self).write(chunk)
                for chunk in appdata:
                    if chunk and not self._closing:
                        self._protocol.data_received(chunk)
                    elif not chunk and self._close_on_unwrap:
                        # close_notify
                        self.close()
        except ssl.SSLError as e:
            self._log.warning('SSL error {} (reason {})', e.errno,
                              getattr(e, 'reason', 'unknown'), exc_info=True)
            self._error = e
            self.abort()
        # Process write backlog. A read could have unblocked a write.
        if not self._error:
            self._process_write_backlog()

    def pause_reading(self):
        """Stop reading data.

        Flow control for SSL is a little bit more complicated than for a
        regular :class:`Transport` because SSL handshakes can occur at any time
        during a connection. These handshakes require reading to be enabled,
        even if the application called :meth:`pause_reading` before.

        The approach taken by Gruvi is that when a handshake occurs, reading is
        always enabled even if :meth:`pause_reading` was called before. I
        believe this is the best way to prevent complex read side buffering
        that could also result in read buffers of arbitrary size.

        The consequence is that if you are implementing your own protocol and
        you want to support SSL, then your protocol should be able to handle a
        callback even if it called :meth:`pause_reading` before.
        """
        if self._sslpipe.need_ssldata:
            return
        super(SslTransport, self).pause_reading()

    def resume_reading(self):
        """Resume reading data.

        See the note in :meth:`pause_reading` for special considerations on
        flow control with SSL.
        """
        return super(SslTransport, self).resume_reading()

    def do_handshake(self):
        """Start the SSL handshake.

        This method only needs to be called if this transport was created with
        *do_handshake_on_connect* set to False (the default is True).

        The handshake needs to be synchronized between the both endpoints, so
        that SSL record level data is not incidentially interpreted as
        plaintext. Usually this is done by starting the handshake directly
        after a connection is established, but you can also use an application
        level protocol.
        """
        if self._error:
            raise compat.saved_exc(self._error)
        elif self._closing or self._handle.closed:
            raise TransportError('SSL transport is closing/closed')
        self._write_backlog.append([b'', True])
        self._process_write_backlog()

    def unwrap(self):
        """Remove the security layer.

        Use this method only if you want to send plaintext data on the
        connection after the security layer has been removed. In all other
        cases, use :meth:`close`.

        If the unwrap is initiated by us, then any data sent after it will be
        buffered until the corresponding close_notify response is received from
        our peer.

        If the unwrap is initiated by the remote peer, then this method will
        acknowledge it. You need an application level protocol to determine
        when to do this because the receipt of a close_notify is not
        communicated to the application.
        """
        if self._error:
            raise compat.saved_exc(self._error)
        elif self._closing or self._handle.closed:
            raise TransportError('SSL transport is closing/closed')
        self._close_on_unwrap = False
        self._write_backlog.append([b'', False])
        self._process_write_backlog()

    def can_write_eof(self):
        # SSL/TLS does not support a half close. Theoretically we could return
        # True here when unwrapped.
        return False

    def close(self):
        """Cleanly shut down the SSL protocol and close the transport."""
        if self._closing or self._handle.closed:
            return
        self._closing = True
        self._write_backlog.append([b'', False])
        self._process_write_backlog()


def create_ssl_context(**sslargs):
    """Create a new SSL context in a way that is compatible with different
    Python versions."""
    context = SSLContext(ssl.PROTOCOL_SSLv23)
    if sslargs.get('certfile'):
        context.load_cert_chain(sslargs['certfile'], sslargs.get('keyfile'))
    if sslargs.get('ca_certs'):
        context.load_verify_locations(sslargs['ca_certs'])
    if sslargs.get('cert_reqs'):
        context.verify_mode = sslargs['cert_reqs']
    if sslargs.get('ciphers'):
        context.set_ciphers(sslargs['ciphers'])
    return context



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2014 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

# Suppress warnings about 'import *' here. The submodules are designed to
# export their symbols to a global package namespace like this.
# flake8: noqa

# should not use "from gruvi import *"
__all__ = []

import sys
if sys.version_info[0] == 2 and sys.version_info[1] < 7 \
        or sys.version_info[0] == 3 and sys.version_info[1] < 3:
    raise ImportError('Gruvi requires Python 2.7 or 3.3+')

# import all the subpackages into the "gruvi" namespace
from .errors import *
from .hub import *
from .fibers import *
from .sync import *
from .local import *
from .ssl import *
from .futures import *
from .transports import *
from .protocols import *
from .endpoints import *
from .address import *
from .process import *
from .stream import *
from .http import *
from .jsonrpc import *
from .dbus import *

# clean up module namespace
del sys, absolute_import, print_function



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2014 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function, division

import random
import unittest

import gruvi
from gruvi import Fiber, Event, Process, Timeout
from gruvi.futures import FiberPool, ThreadPool, Future
from support import UnitTest


class TestFuture(UnitTest):

    def test_result(self):
        fut = Future()
        fut.set_result(10)
        self.assertEqual(fut.result(), 10)
        self.assertIsNone(fut.exception())
        self.assertEqual(fut.result(), 10)

    def test_exception(self):
        fut = Future()
        fut.set_exception(ValueError())
        self.assertIsInstance(fut.exception(), ValueError)
        self.assertRaises(ValueError, fut.result)

    def test_wait_result(self):
        fut = Future()
        def set_result():
            gruvi.sleep(0.1)
            fut.set_result('foo')
        fib = gruvi.Fiber(set_result)
        fib.start()
        self.assertEqual(fut.result(), 'foo')

    def test_wait_result_timeout(self):
        fut = Future()
        self.assertRaises(gruvi.Timeout, fut.result, 0.01)

    def test_wait_exception(self):
        fut = Future()
        def set_result():
            gruvi.sleep(0.1)
            fut.set_exception(RuntimeError)
        fib = gruvi.Fiber(set_result)
        fib.start()
        self.assertEqual(fut.exception(), RuntimeError)

    def test_wait_exception_timeout(self):
        fut = Future()
        self.assertRaises(gruvi.Timeout, fut.exception, 0.01)

    def test_set_running(self):
        fut = Future()
        self.assertFalse(fut.running())
        fut.set_running()
        self.assertTrue(fut.running())

    def test_cancel(self):
        fut = Future()
        self.assertFalse(fut.cancelled())
        self.assertTrue(fut.cancel())
        self.assertTrue(fut.cancelled())

    def test_cancel_running(self):
        fut = Future()
        self.assertFalse(fut.cancelled())
        fut.set_running(True)
        self.assertTrue(fut.running())
        self.assertTrue(fut.cancel())
        self.assertTrue(fut.cancelled())

    def test_cancel_running_nocancel(self):
        fut = Future()
        self.assertFalse(fut.cancelled())
        fut.set_running(False)
        self.assertTrue(fut.running())
        self.assertFalse(fut.cancel())
        self.assertFalse(fut.cancelled())

    def test_set_result_twice(self):
        fut = Future()
        fut.set_result('foo')
        self.assertEqual(fut.result(), 'foo')
        fut.set_result('bar')
        self.assertEqual(fut.result(), 'foo')

    def test_set_exception_twice(self):
        fut = Future()
        fut.set_exception(RuntimeError)
        self.assertEqual(fut.exception(), RuntimeError)
        fut.set_exception(ValueError)
        self.assertEqual(fut.exception(), RuntimeError)

    def test_set_running_twice(self):
        fut = Future()
        fut.set_running(False)
        state = fut._state
        fut.set_running(True)
        self.assertEqual(state, fut._state)

    def test_callbacks(self):
        cbargs = []
        def callback(*args):
            cbargs.append(args)
        fut = Future()
        fut.add_done_callback(callback)
        fut.add_done_callback(callback, 'foo')
        fut.add_done_callback(callback, 'foo', 'bar')
        fut.add_done_callback(callback, fut)
        handle = fut.add_done_callback(callback, 'baz')
        self.assertIsNotNone(handle)
        fut.remove_done_callback(handle)
        fut.set_result('quz')
        self.assertEqual(len(cbargs), 4)
        self.assertEqual(cbargs[0], ())
        self.assertEqual(cbargs[1], ('foo',))
        self.assertEqual(cbargs[2], ('foo', 'bar'))
        self.assertEqual(cbargs[3], (fut,))


class PoolTest(object):

    def setUp(self):
        self.pool = self.Pool(self.maxsize)

    def tearDown(self):
        self.pool.close()

    def test_simple(self):
        def func(val):
            return val
        fut = self.pool.submit(func, 'foo')
        self.assertIsInstance(fut, Future)
        self.assertEqual(fut.result(), 'foo')

    def test_submit_sleep(self):
        def func(val):
            gruvi.sleep(0.1)
            return val
        fut = self.pool.submit(func, 'foo')
        self.assertEqual(fut.result(), 'foo')

    def test_submit_many(self):
        def func(val):
            return val
        futures = []
        for i in range(self.count):
            futures.append(self.pool.submit(func, i))
        result = []
        for fut in futures:
            result.append(fut.result())
        result.sort()
        self.assertEqual(result, list(range(self.count)))

    def test_submit_many_sleep(self):
        def func(val):
            gruvi.sleep(0.01)
            return val
        futures = []
        for i in range(self.count):
            futures.append(self.pool.submit(func, i))
        result = []
        for fut in futures:
            result.append(fut.result())
        result.sort()
        self.assertEqual(result, list(range(self.count)))

    def test_submit_exception(self):
        def func():
            raise ValueError()
        fut = self.pool.submit(func)
        self.assertRaises(ValueError, fut.result)

    def test_map(self):
        def double(x):
            return x*2
        result = self.pool.map(double, range(self.count))
        self.assertEqual(list(result), list(range(0, 2*self.count, 2)))

    def test_map_order(self):
        # Make sure that map() respects the order of the input. The random
        # delay in double() will make the results ready out of order, but the
        # return value of map() should be in order nonetheless.
        def double(x):
            gruvi.sleep(random.randint(0, 2) / 1000)
            return x*2
        result = self.pool.map(double, range(self.count))
        self.assertEqual(list(result), list(range(0, 2*self.count, 2)))

    def test_map_timeout(self):
        # Ensure that the "timeout" argument to map() works.
        result = self.pool.map(gruvi.sleep, (0.2,), timeout=0.1)
        self.assertRaises(gruvi.Timeout, list, result)

    def test_pool_close(self):
        def func(i):
            return i
        pool = self.Pool()
        futures = []
        for i in range(self.count):
            futures.append(pool.submit(func, i))
        result = [fut.result() for fut in futures]
        result.sort()
        self.assertEqual(result, list(range(self.count)))
        self.assertGreater(len(pool._workers), 0)
        pool.close()
        self.assertEqual(len(pool._workers), 0)
        self.assertRaises(RuntimeError, pool.submit, func, i)
        self.assertRaises(RuntimeError, list, pool.map(func, [i]))


class TestFiberPool(PoolTest, UnitTest):

    maxsize = 50
    count = 100
    Pool = FiberPool


class TestThreadPool(PoolTest, UnitTest):

    maxsize = 5
    count = 100
    Pool = ThreadPool


class TestWait(UnitTest):

    def test_basic(self):
        # Simple wait() test waiting for two futures.
        futures = [Future() for i in range(2)]
        result = []
        def waiter():
            result.append(gruvi.wait(futures))
        fib = Fiber(waiter)
        fib.start()
        gruvi.sleep(0)
        self.assertTrue(fib.alive)
        self.assertEqual(result, [])
        futures[0].set_result(None)
        gruvi.sleep(0)
        self.assertTrue(fib.alive)
        futures[1].set_result(None)
        gruvi.sleep(0)
        self.assertFalse(fib.alive)
        self.assertEqual(result, [(futures, [])])

    def test_partial(self):
        # Ensure that the "count" argument of wait() can be used to wait for a
        # subset of the futures to complete.
        futures = [Future() for i in range(10)]
        result = []
        def waiter():
            result.append(gruvi.wait(futures, count=5))
        fib = Fiber(waiter)
        fib.start()
        gruvi.sleep(0)
        for i in range(4):
            futures[i].set_result(None)
        self.assertTrue(fib.alive)
        self.assertEqual(result, [])
        futures[4].set_result(None)
        gruvi.sleep(0)
        self.assertFalse(fib.alive)
        self.assertEqual(result, [(futures[:5], futures[5:])])

    def test_partial_fibers(self):
        # Have a lot of fibers wait on a small subet of a lot of futures.
        futures = [Future() for i in range(100)]
        def waiter():
            gruvi.wait(futures, count=2)
        fibers = [Fiber(waiter) for i in range(100)]
        for fib in fibers:
            fib.start()
        gruvi.sleep(0)
        futures[0].set_result(None)
        gruvi.sleep(0)
        for fib in fibers:
            self.assertTrue(fib.alive)
        futures[len(futures)//2].set_result(None)
        gruvi.sleep(0)
        for fib in fibers:
            self.assertFalse(fib.alive)
        # Test cleanup
        for fut in futures:
            self.assertFalse(fut._callbacks)

    def test_timeout(self):
        futures = [Future() for i in range(10)]
        result = gruvi.wait(futures, timeout=0.01)
        self.assertEqual(result, ([], futures))
        for fut in futures:
            self.assertFalse(fut._callbacks)

    def test_timeout_fibers(self):
        futures = [Future() for i in range(10)]
        result = []
        def waiter():
            result.append(gruvi.wait(futures, timeout=0.01))
        fibers = [Fiber(waiter) for i in range(10)]
        for fib in fibers:
            fib.start()
        for fib in fibers:
            fib.join()
        self.assertEqual(len(result), 10)
        for res in result:
            self.assertEqual(res, ([], futures))
        for fut in futures:
            self.assertFalse(fut._callbacks)

    def test_duplicate(self):
        futures = [Future()] * 5
        result = []
        def waiter():
            result.append(gruvi.wait(futures))
        fib = Fiber(waiter)
        fib.start()
        gruvi.sleep(0)
        self.assertEqual(result, [])
        futures[0].set_result(None)
        gruvi.sleep(0)
        self.assertEqual(result, [(futures, [])])

    def test_object_types(self):
        def sleeper():
            gruvi.sleep(0.1)
        objects = []
        for i in range(2):
            fib = Fiber(sleeper)
            fib.start()
            objects.append(fib)
        for i in range(2):
            objects.append(Event())
        for i in range(2):
            objects.append(Future())
        for i in range(2):
            proc = Process()
            proc.spawn('true')
            objects.append(proc)
        result = []
        def waiter():
            result.append(gruvi.wait(objects))
        fib = Fiber(waiter)
        fib.start()
        gruvi.sleep(0)
        self.assertTrue(fib.alive)
        objects[2].set()
        objects[3].set()
        objects[4].set_result(True)
        objects[5].set_exception(RuntimeError)
        fib.join()
        self.assertFalse(fib.alive)
        self.assertEqual(set(result[0][0]), set(objects))
        self.assertFalse(objects[0].alive)
        self.assertFalse(objects[1].alive)
        self.assertEqual(objects[6].returncode, 0)
        self.assertEqual(objects[7].returncode, 0)
        for obj in objects:
            self.assertFalse(obj._callbacks)
        objects[6].close()
        objects[7].close()


class TestAsCompleted(UnitTest):

    def test_order(self):
        futures = [Future() for i in range(100)]
        order = list(range(len(futures)))
        random.shuffle(order)
        futures[order[0]].set_result(None)
        count = 0
        for fut in gruvi.as_completed(futures):
            self.assertEqual(fut, futures[order[count]])
            count += 1
            if count < len(order):
                futures[order[count]].set_result(None)
        self.assertEqual(count, len(futures))
        for fut in futures:
            self.assertFalse(fut._callbacks)

    def test_partial(self):
        futures = [Future() for i in range(100)]
        futures[0].set_result(None)
        count = 0
        for fut in gruvi.as_completed(futures, 10):
            self.assertEqual(fut, futures[count])
            count += 1
            futures[count].set_result(None)
        self.assertEqual(count, 10)
        for fut in futures:
            self.assertFalse(fut._callbacks)

    def test_partial_fibers(self):
        futures = [Future() for i in range(100)]
        result = []
        def waiter():
            for fut in gruvi.as_completed(futures, count=2):
                result.append(fut)
        fibers = [Fiber(waiter) for i in range(100)]
        for fib in fibers:
            fib.start()
        gruvi.sleep(0)
        futures[0].set_result(None)
        for fib in fibers:
            self.assertTrue(fib.alive)
        futures[len(futures)//2].set_result(None)
        gruvi.sleep(0)
        for fib in fibers:
            self.assertFalse(fib.alive)
        self.assertEqual(len(result), len(fibers)*2)
        for fut in futures:
            self.assertFalse(fut._callbacks)

    def test_timeout(self):
        futures = [Future() for i in range(10)]
        self.assertRaises(Timeout, lambda: list(gruvi.as_completed(futures, timeout=0.01)))
        for fut in futures:
            self.assertFalse(fut._callbacks)

    def test_timeout_fibers(self):
        futures = [Future() for i in range(10)]
        exceptions = []
        def waiter():
            try:
                list(gruvi.as_completed(futures, timeout=0.01))
            except Timeout as e:
                exceptions.append(e)
        fibers = [Fiber(waiter) for i in range(10)]
        for fib in fibers:
            fib.start()
        for fib in fibers:
            fib.join()
        self.assertEqual(len(exceptions), 10)
        for exc in exceptions:
            self.assertIsInstance(exc, Timeout)
        for fut in futures:
            self.assertFalse(fut._callbacks)


if __name__ == '__main__':
    unittest.main()



"""
Module to represent DBus Messages

@author: Tom Cocagne
"""
from __future__ import absolute_import, print_function

from . import marshal, error


_headerFormat = 'yyyyuua(yv)'

    
class DBusMessage (object):
    """
    Abstract base class for DBus messages

    @ivar _messageType: C{int} DBus message type
    @ivar expectReply: True if a method return message is expected
    @ivar autoStart: True if a service should be auto started by this message
    @ivar signature: C{str} DBus signature describing the body content
    @ivar endian: C{int} containing endian code: Little endian = ord('l'). Big
                  endian is ord('B'). Defaults to little-endian
    @ivar bodyLength: Length of the body in bytes
    @ivar serial: C{int} message serial number
    @ivar rawMessage: Raw binary message data
    @ivar rawHeader: Raw binary message header
    @ivar rawBody: Raw binary message body
    @ivar interface: C{str} DBus interface name
    @ivar path: C{str} DBus object path
    @ivar sender: C{str} DBus bus name for sending connection
    @ivar destination: C{str} DBus bus name for destination connection
    
    """
    _maxMsgLen         = 2**27
    _nextSerial        = 1
    _protocolVersion   = 1

    # Overriden by subclasses
    _messageType       = 0    
    _headerAttrs       = None # [(attr_name, code, is_required), ...]

    # Set prior to marshal or during/after unmarshalling
    expectReply        = True 
    autoStart          = True
    signature          = None
    body               = None

    # Set during marshalling/unmarshalling
    endian             = ord('l')
    bodyLength         = 0     
    serial             = None
    headers            = None
    rawMessage         = None

    # Required/Optional
    interface          = None
    path               = None
    
    # optional
    sender             = None
    destination        = None


#    def printSelf(self):
#        mtype = { 1 : 'MethodCall',
#                  2 : 'MethodReturn',
#                  3 : 'Error',
#                  4 : 'Signal' }
#        print mtype[self._messageType]
#        keys = self.__dict__.keys()
#        keys.sort()
#        for a in keys:
#            if not a.startswith('raw'):
#                print '    %s = %s' % (a.ljust(15), str(getattr(self,a)))

    
    def _marshal(self, newSerial=True):
        """
        Encodes the message into binary format. The resulting binary message is
        stored in C{self.rawMessage}
        """
        flags = 0

        if not self.expectReply:
            flags |= 0x1

        if not self.autoStart:
            flags |= 0x2
        
        self.headers = list()
        
        for attr_name, code, is_required in self._headerAttrs:
            hval = getattr(self, attr_name, None)
            
            if hval is not None:
                if attr_name == 'path':
                    hval = marshal.ObjectPath(hval)
                elif attr_name == 'signature':
                    hval = marshal.Signature(hval)
                    
                self.headers.append( [code, hval] )

        if self.signature:
            binBody = b''.join( marshal.marshal( self.signature, self.body )[1] )
        else:
            binBody = b''

        self.bodyLength = len(binBody)

        if newSerial:
            self.serial = DBusMessage._nextSerial

            DBusMessage._nextSerial += 1
        
        binHeader = b''.join(marshal.marshal(_headerFormat,
                                            [self.endian,
                                             self._messageType,
                                             flags,
                                             self._protocolVersion,
                                             self.bodyLength,
                                             self.serial,
                                             self.headers],
                                            lendian = self.endian == ord('l') )[1])
        
        headerPadding = marshal.pad['header']( len(binHeader) )

        self.rawHeader  = binHeader
        self.rawPadding = headerPadding
        self.rawBody    = binBody
        
        self.rawMessage = b''.join( [binHeader, headerPadding, binBody] )

        if len(self.rawMessage) > self._maxMsgLen:
            raise error.MarshallingError('Marshalled message exceeds maximum message size of %d' %
                                         (self._maxMsgLen,))



class MethodCallMessage (DBusMessage):
    """
    A DBus Method Call Message
    """
    _messageType = 1
    _headerAttrs = [ ('path',        1, True ),
                     ('interface',   2, False),
                     ('member',      3, True ),
                     ('destination', 6, False),
                     ('sender',      7, False),
                     ('signature',   8, False) ]


    def __init__(self, path, member, interface=None, destination=None,
                 signature=None, body=None,
                 expectReply=True, autoStart=True):
        """
        @param path: C{str} DBus object path
        @param member: C{str} Member name
        @param interface: C{str} DBus interface name or None
        @param destination: C{str} DBus bus name for message destination or
                            None
        @param signature: C{str} DBus signature string for encoding
                          C{self.body}
        @param body: C{list} of python objects to encode. Objects must match
                     the C{self.signature}
        @param expectReply: True if a Method Return message should be sent
                            in reply to this message
        @param autoStart: True if the Bus should auto-start a service to handle
                          this message if the service is not already running.
        """
        
        marshal.validateMemberName( member )
        
        if interface:
            marshal.validateInterfaceName(interface)
            
        if destination:
            marshal.validateBusName(destination)

        if path == '/org/freedesktop/DBus/Local':
            raise error.MarshallingError('/org/freedesktop/DBus/Local is a reserved path')
            
        self.path         = path
        self.member       = member
        self.interface    = interface
        self.destination  = destination
        self.signature    = signature
        self.body         = body
        self.expectReply  = expectReply
        self.autoStart    = autoStart

        self._marshal()
        


class MethodReturnMessage (DBusMessage):
    """
    A DBus Method Return Message
    """
    _messageType = 2
    _headerAttrs = [ ('reply_serial', 5, True ),
                     ('destination',  6, False),
                     ('sender',       7, False),
                     ('signature',    8, False) ]

    def __init__(self, reply_serial, body=None,
                 destination=None, signature=None):
        """
        @param reply_serial: C{int} serial number this message is a reply to
        @param destination: C{str} DBus bus name for message destination or
                            None
        @param signature: C{str} DBus signature string for encoding
                          C{self.body}
        @param body: C{list} of python objects to encode. Objects must match
                     the C{self.signature}
        """
        if destination:
            marshal.validateBusName(destination)

        self.reply_serial = marshal.UInt32(reply_serial)
        self.destination  = destination
        self.signature    = signature
        self.body         = body

        self._marshal()


class ErrorMessage (DBusMessage):
    """
    A DBus Error Message
    """
    _messageType = 3
    _headerAttrs = [ ('error_name',   4, True ),
                     ('reply_serial', 5, True ),
                     ('destination',  6, False),
                     ('sender',       7, False),
                     ('signature',    8, False) ]

    def __init__(self, error_name, reply_serial, destination=None, signature=None,
                 body=None, sender=None):
        """
        @param error_name: C{str} DBus error name
        @param reply_serial: C{int} serial number this message is a reply to
        @param destination: C{str} DBus bus name for message destination or
                            None
        @param signature: C{str} DBus signature string for encoding
                          C{self.body}
        @param body: C{list} of python objects to encode. Objects must match
                     the C{self.signature}
        @param sender: C{str} name of the originating Bus connection
        """
        if destination:
            marshal.validateBusName(destination)

        marshal.validateInterfaceName(error_name)

        self.error_name   = error_name
        self.reply_serial = marshal.UInt32(reply_serial)
        self.destination  = destination
        self.signature    = signature
        self.body         = body
        self.sender       = sender

        self._marshal()


        
class SignalMessage (DBusMessage):
    """
    A DBus Signal Message
    """
    _messageType = 4
    _headerAttrs = [ ('path',        1, True ),
                     ('interface',   2, True ),
                     ('member',      3, True ),
                     ('destination', 6, False),
                     ('sender',      7, False),
                     ('signature',   8, False) ]

    def __init__(self, path, member, interface, destination=None, signature=None,
                 body=None):
        """
        @param path: C{str} DBus object path of the object sending the signal
        @param member: C{str} Member name
        @param interface: C{str} DBus interface name or None
        @param destination: C{str} DBus bus name for message destination or
                            None
        @param signature: C{str} DBus signature string for encoding
                          C{self.body}
        @param body: C{list} of python objects to encode. Objects must match
                     the C{self.signature}
        """
        marshal.validateMemberName( member )
        marshal.validateInterfaceName(interface)
            
        if destination:
            marshal.validateBusName(destination)
            
        self.path        = path
        self.member      = member
        self.interface   = interface
        self.destination = destination
        self.signature   = signature
        self.body        = body

        self._marshal()


_mtype = { 1 : MethodCallMessage,
           2 : MethodReturnMessage,
           3 : ErrorMessage,
           4 : SignalMessage }

_hcode = { 1 : 'path',
           2 : 'interface',
           3 : 'member',
           4 : 'error_name',
           5 : 'reply_serial',
           6 : 'destination',
           7 : 'sender',
           8 : 'signature',
           9 : 'unix_fds' }


def parseMessage( rawMessage ):
    """
    Parses the raw binary message and returns a L{DBusMessage} subclass

    @type rawMessage: C{str}
    @param rawMessage: Raw binary message to parse

    @rtype: L{DBusMessage} subclass
    @returns: The L{DBusMessage} subclass corresponding to the contained
              message
    """

    lendian = rawMessage[0] == b'l'[0]
    
    nheader, hval = marshal.unmarshal(_headerFormat, rawMessage, 0, lendian)

    messageType = hval[1]

    if not messageType in _mtype:
        raise error.MarshallingError('Unknown Message Type: ' + str(messageType))

    m = object.__new__( _mtype[messageType] )

    m.rawHeader = rawMessage[:nheader]

    npad = nheader % 8 and (8 - nheader%8) or 0

    m.rawPadding = rawMessage[nheader: nheader+npad]

    m.rawBody = rawMessage[ nheader + npad: ]

    m.serial = hval[5]
    
    for code, v in hval[6]:
        try:
            setattr(m, _hcode[code], v)
        except KeyError:
            pass
        
    if m.signature:
        nbytes, m.body = marshal.unmarshal(m.signature, m.rawBody, lendian = lendian)

    return m

    
        
        
        
        



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2014 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import six
import unittest

from gruvi.sslcompat import MemoryBIO
from support import UnitTest


class MemoryBIOTests(UnitTest):

    def test_read_write(self):
        bio = MemoryBIO()
        bio.write(b'foo')
        self.assertEqual(bio.read(), b'foo')
        self.assertEqual(bio.read(), b'')
        bio.write(b'foo')
        bio.write(b'bar')
        self.assertEqual(bio.read(), b'foobar')
        self.assertEqual(bio.read(), b'')
        bio.write(b'baz')
        self.assertEqual(bio.read(2), b'ba')
        self.assertEqual(bio.read(1), b'z')
        self.assertEqual(bio.read(1), b'')

    def test_eof(self):
        bio = MemoryBIO()
        self.assertFalse(bio.eof)
        self.assertEqual(bio.read(), b'')
        self.assertFalse(bio.eof)
        bio.write(b'foo')
        self.assertFalse(bio.eof)
        bio.write_eof()
        self.assertFalse(bio.eof)
        self.assertEqual(bio.read(2), b'fo')
        self.assertFalse(bio.eof)
        self.assertEqual(bio.read(1), b'o')
        self.assertTrue(bio.eof)
        self.assertEqual(bio.read(), b'')
        self.assertTrue(bio.eof)

    def test_pending(self):
        bio = MemoryBIO()
        self.assertEqual(bio.pending, 0)
        bio.write(b'foo')
        self.assertEqual(bio.pending, 3)
        for i in range(3):
            bio.read(1)
            self.assertEqual(bio.pending, 3-i-1)
        for i in range(3):
            bio.write(b'x')
            self.assertEqual(bio.pending, i+1)
        bio.read()
        self.assertEqual(bio.pending, 0)

    def test_buffer_types(self):
        bio = MemoryBIO()
        bio.write(b'foo')
        self.assertEqual(bio.read(), b'foo')
        bio.write(bytearray(b'bar'))
        self.assertEqual(bio.read(), b'bar')
        bio.write(memoryview(b'baz'))
        self.assertEqual(bio.read(), b'baz')

    def test_error_types(self):
        bio = MemoryBIO()
        if six.PY3:
            self.assertRaises(TypeError, bio.write, 'foo')
        self.assertRaises(TypeError, bio.write, None)
        self.assertRaises(TypeError, bio.write, True)
        self.assertRaises(TypeError, bio.write, 1)


if __name__ == '__main__':
    unittest.main()



#
# This file is part of gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2017 the gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import unittest

import gruvi
from support import UnitTest


class TestFiber(UnitTest):

    def test_basic(self):
        # Spawn fibers and join them.
        counter = [0]
        def worker():
            counter[0] += 1
            gruvi.sleep(0.1)
        fibers = []
        for i in range(1000):
            fibers.append(gruvi.spawn(worker))
        for fiber in fibers:
            self.assertTrue(fiber.alive)
        for fiber in fibers:
            fiber.join()
        self.assertEqual(counter[0], 1000)
        for fiber in fibers:
            self.assertFalse(fiber.alive)

    def test_spawn_nested(self):
        # Spawn fibers from fiber.
        joined = []
        def spawn_fiber(i):
            if i == 0:
                return
            fiber = gruvi.spawn(spawn_fiber, i-1)
            fiber.join()
            joined.append(i-1)
        fiber = gruvi.spawn(spawn_fiber, 1000)
        fiber.join()
        self.assertEqual(joined, list(range(1000)))

    def test_switch_hub_only(self):
        # Ensure that only the hub may switch to or throw into a fiber.
        ready = []
        exceptions = []
        def worker():
            ready.append(gruvi.current_fiber())
            try:
                gruvi.sleep(0.1)
            except Exception as e:
                exceptions.append(e)
        fiber = gruvi.Fiber(worker)
        self.assertRaises(RuntimeError, fiber.switch)
        self.assertRaises(RuntimeError, fiber.throw, None)
        hub = gruvi.get_hub()
        hub.run_callback(fiber.switch)
        gruvi.sleep(0)
        self.assertEqual(ready, [fiber])
        error = RuntimeError('foo')
        hub.run_callback(fiber.throw, RuntimeError, error)
        gruvi.sleep(0)
        self.assertEqual(exceptions, [error])

    def test_cancel(self):
        # Ensure that it's possible to cancel() a fiber.
        def sleeper():
            gruvi.sleep(1000)
        fiber = gruvi.spawn(sleeper)
        gruvi.sleep(0)
        self.assertTrue(fiber.alive)
        fiber.cancel()
        gruvi.sleep(0)
        self.assertFalse(fiber.alive)


if __name__ == '__main__':
    unittest.main()



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2014 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import six
import pyuv

from .hub import get_hub, switch_back, switchpoint

__all__ = ['saddr', 'paddr', 'getaddrinfo', 'getnameinfo']


def saddr(address):
    """Return a string representation for an address.

    The *address* paramater can be a pipe name, an IP address tuple, or a
    socket address.

    The return value is always a ``str`` instance.
    """
    if isinstance(address, six.binary_type) and six.PY3:
        return address.decode('utf8')
    elif isinstance(address, six.string_types):
        return address
    elif isinstance(address, tuple) and ':' in address[0]:
        return '[{}]:{}'.format(address[0], address[1])
    elif isinstance(address, tuple):
        return '{}:{}'.format(*address)
    else:
        raise TypeError('illegal address type: {!s}'.format(type(address)))


def paddr(address):
    """Parse a string representation of an address.

    This function is the inverse of :func:`saddr`.
    """
    if address.startswith('['):
        p1 = address.find(']:')
        if p1 == -1:
            raise ValueError
        return (address[1:p1], int(address[p1+2:]))
    elif ':' in address:
        p1 = address.find(':')
        return (address[:p1], int(address[p1+1:]))
    else:
        return address


@switchpoint
def getaddrinfo(node, service=0, family=0, socktype=0, protocol=0, flags=0, timeout=30):
    """Resolve an Internet *node* name and *service* into a socket address.

    The *family*, *socktype* and *protocol* are optional arguments that specify
    the address family, socket type and protocol, respectively. The *flags*
    argument allows you to pass flags to further modify the resolution process.
    See the :func:`socket.getaddrinfo` function for a detailed description of
    these arguments.

    The return value is a list of ``(family, socktype, proto, canonname,
    sockaddr)`` tuples. The fifth element (``sockaddr``) is the socket address.
    It will be a 2-tuple ``(addr, port)`` for an IPv4 address, and a 4-tuple
    ``(addr, port, flowinfo, scopeid)`` for an IPv6 address.

    The address resolution is performed in the libuv thread pool.
    """
    hub = get_hub()
    with switch_back(timeout) as switcher:
        request = pyuv.dns.getaddrinfo(hub.loop, node, service, family,
                                       socktype, protocol, flags, callback=switcher)
        switcher.add_cleanup(request.cancel)
        result = hub.switch()
    result, error = result[0]
    if error:
        message = pyuv.errno.strerror(error)
        raise pyuv.error.UVError(error, message)
    return result


@switchpoint
def getnameinfo(sockaddr, flags=0, timeout=30):
    """Resolve a socket address *sockaddr* back to a ``(node, service)`` tuple.

    The *flags* argument can be used to modify the resolution process. See the
    :func:`socket.getnameinfo` function for more information.

    The address resolution is performed in the libuv thread pool.
    """
    hub = get_hub()
    with switch_back(timeout) as switcher:
        request = pyuv.dns.getnameinfo(hub.loop, sockaddr, flags, callback=switcher)
        switcher.add_cleanup(request.cancel)
        result = hub.switch()
    result, error = result[0]
    if error:
        message = pyuv.errno.strerror(error)
        raise pyuv.error.UVError(error, message)
    return result



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2017 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import os
import six
import hashlib
import unittest
from io import TextIOWrapper

import gruvi
from gruvi.stream import Stream, StreamProtocol, StreamClient, StreamServer
from gruvi.errors import Timeout
from gruvi.transports import TransportError
from support import UnitTest, MockTransport


class TestStream(UnitTest):

    def test_read(self):
        stream = Stream(None)
        stream.buffer.feed(b'foo')
        self.assertEqual(stream.read(3), b'foo')
        stream.buffer.feed(b'foo bar')
        self.assertEqual(stream.read(3), b'foo')
        self.assertEqual(stream.read(4), b' bar')

    def test_read_incr(self):
        stream = Stream(None)
        buf = b'foobar'
        for i in range(len(buf)):
            stream.buffer.feed(buf[i:i+1])
        stream.buffer.feed_eof()
        self.assertEqual(stream.read(), b'foobar')

    def test_read_eof(self):
        stream = Stream(None)
        stream.buffer.feed(b'foo')
        stream.buffer.feed_eof()
        self.assertEqual(stream.read(), b'foo')
        self.assertEqual(stream.read(), b'')

    def test_read_wait(self):
        stream = Stream(None)
        stream.buffer.feed(b'foo')
        def write_more():
            gruvi.sleep(0.01)
            stream.buffer.feed(b'bar')
            gruvi.sleep(0.01)
            stream.buffer.feed_eof()
        gruvi.spawn(write_more)
        self.assertEqual(stream.read(), b'foobar')
        self.assertEqual(stream.read(), b'')

    def test_read_error(self):
        stream = Stream(None)
        stream.buffer.feed(b'foo')
        stream.buffer.feed_error(RuntimeError)
        self.assertEqual(stream.read(), b'foo')
        self.assertRaises(RuntimeError, stream.read)

    def test_read_wait_error(self):
        stream = Stream(None)
        stream.buffer.feed(b'foo')
        def write_more():
            gruvi.sleep(0.01)
            stream.buffer.feed(b'bar')
            gruvi.sleep(0.01)
            stream.buffer.feed_error(RuntimeError)
        gruvi.spawn(write_more)
        self.assertEqual(stream.read(), b'foobar')
        self.assertRaises(RuntimeError, stream.read)

    def test_read_eof_and_error(self):
        stream = Stream(None)
        stream.buffer.feed(b'foo')
        stream.buffer.feed_eof()
        stream.buffer.feed_error(RuntimeError)
        self.assertEqual(stream.read(), b'foo')
        self.assertEqual(stream.read(), b'')

    def test_read1(self):
        stream = Stream(None)
        stream.buffer.feed(b'foobar')
        self.assertEqual(stream.read1(3), b'foo')
        self.assertEqual(stream.read1(100), b'bar')

    def test_read1_wait(self):
        stream = Stream(None)
        stream.buffer.feed(b'foo')
        def write_more():
            gruvi.sleep(0.01)
            stream.buffer.feed(b'bar')
            gruvi.sleep(0.01)
            stream.buffer.feed_eof()
        gruvi.spawn(write_more)
        self.assertEqual(stream.read1(100), b'foo')
        self.assertEqual(stream.read1(100), b'bar')
        self.assertEqual(stream.read1(100), b'')

    def test_readline(self):
        stream = Stream(None)
        stream.buffer.feed(b'foo\n')
        self.assertEqual(stream.readline(), b'foo\n')
        stream.buffer.feed(b'bar\nbaz\n')
        self.assertEqual(stream.readline(), b'bar\n')
        self.assertEqual(stream.readline(), b'baz\n')

    def test_readline_incr(self):
        stream = Stream(None)
        buf = b'foo\nbar\n'
        for i in range(len(buf)):
            stream.buffer.feed(buf[i:i+1])
        self.assertEqual(stream.readline(), b'foo\n')
        self.assertEqual(stream.readline(), b'bar\n')

    def test_readline_limit(self):
        stream = Stream(None)
        stream.buffer.feed(b'foobar\n')
        self.assertEqual(stream.readline(3), b'foo')
        self.assertEqual(stream.readline(), b'bar\n')

    def test_readline_eof(self):
        stream = Stream(None)
        stream.buffer.feed(b'foo')
        stream.buffer.feed_eof()
        self.assertEqual(stream.readline(), b'foo')
        self.assertEqual(stream.readline(), b'')

    def test_readline_wait_eof(self):
        stream = Stream(None)
        stream.buffer.feed(b'foo')
        def write_more():
            gruvi.sleep(0.01)
            stream.buffer.feed(b'bar')
            gruvi.sleep(0.01)
            stream.buffer.feed_eof()
        gruvi.spawn(write_more)
        self.assertEqual(stream.readline(), b'foobar')
        self.assertEqual(stream.readline(), b'')

    def test_readline_error(self):
        stream = Stream(None)
        stream.buffer.feed(b'foo')
        stream.buffer.feed_error(RuntimeError)
        self.assertEqual(stream.readline(), b'foo')
        self.assertRaises(RuntimeError, stream.readline)

    def test_readline_wait_error(self):
        stream = Stream(None)
        stream.buffer.feed(b'foo')
        def write_more():
            gruvi.sleep(0.01)
            stream.buffer.feed(b'bar')
            gruvi.sleep(0.01)
            stream.buffer.feed_error(RuntimeError)
        gruvi.spawn(write_more)
        self.assertEqual(stream.readline(), b'foobar')
        self.assertRaises(RuntimeError, stream.readline)

    def test_readline_eof_and_error(self):
        stream = Stream(None)
        stream.buffer.feed(b'foo')
        stream.buffer.feed_eof()
        stream.buffer.feed_error(RuntimeError)
        self.assertEqual(stream.readline(), b'foo')
        self.assertEqual(stream.readline(), b'')

    def test_readlines_limit(self):
        stream = Stream(None)
        stream.buffer.feed(b'foo\nbar\n')
        self.assertEqual(stream.readlines(4), [b'foo\n', b'bar\n'])

    def test_readlines_eof(self):
        stream = Stream(None)
        stream.buffer.feed(b'foo\nbar\n')
        stream.buffer.feed_eof()
        self.assertEqual(stream.readlines(), [b'foo\n', b'bar\n'])
        self.assertEqual(stream.readlines(), [])

    def test_readlines_wait_eof(self):
        stream = Stream(None)
        stream.buffer.feed(b'foo\n')
        def write_more():
            gruvi.sleep(0.01)
            stream.buffer.feed(b'bar\n')
            gruvi.sleep(0.01)
            stream.buffer.feed_eof()
        gruvi.spawn(write_more)
        self.assertEqual(stream.readlines(), [b'foo\n', b'bar\n'])
        self.assertEqual(stream.readlines(), [])

    def test_readlines_error(self):
        stream = Stream(None)
        stream.buffer.feed(b'foo\nbar\n')
        stream.buffer.feed_error(RuntimeError)
        self.assertEqual(stream.readlines(), [b'foo\n', b'bar\n'])
        self.assertRaises(RuntimeError, stream.readlines)

    def test_readlines_wait_error(self):
        stream = Stream(None)
        stream.buffer.feed(b'foo\n')
        def write_more():
            gruvi.sleep(0.01)
            stream.buffer.feed(b'bar\n')
            gruvi.sleep(0.01)
            stream.buffer.feed_error(RuntimeError)
        gruvi.spawn(write_more)
        self.assertEqual(stream.readlines(), [b'foo\n', b'bar\n'])
        self.assertRaises(RuntimeError, stream.readlines)

    def test_readlines_eof_and_error(self):
        stream = Stream(None)
        stream.buffer.feed(b'foo\nbar\n')
        stream.buffer.feed_eof()
        stream.buffer.feed_error(RuntimeError)
        self.assertEqual(stream.readlines(), [b'foo\n', b'bar\n'])
        self.assertEqual(stream.readlines(), [])

    def test_iter_eof(self):
        stream = Stream(None)
        stream.buffer.feed(b'foo\nbar\n')
        stream.buffer.feed_eof()
        it = iter(stream)
        self.assertEqual(six.next(it), b'foo\n')
        self.assertEqual(six.next(it), b'bar\n')
        self.assertRaises(StopIteration, six.next, it)

    def test_iter_wait_eof(self):
        stream = Stream(None)
        stream.buffer.feed(b'foo\n')
        def write_more():
            gruvi.sleep(0.01)
            stream.buffer.feed(b'bar\n')
            gruvi.sleep(0.01)
            stream.buffer.feed_eof()
        gruvi.spawn(write_more)
        it = iter(stream)
        self.assertEqual(six.next(it), b'foo\n')
        self.assertEqual(six.next(it), b'bar\n')
        self.assertRaises(StopIteration, six.next, it)

    def test_iter_error(self):
        stream = Stream(None)
        stream.buffer.feed(b'foo\nbar\n')
        stream.buffer.feed_error(RuntimeError)
        it = iter(stream)
        self.assertEqual(six.next(it), b'foo\n')
        self.assertEqual(six.next(it), b'bar\n')
        self.assertRaises(RuntimeError, six.next, it)

    def test_iter_wait_error(self):
        stream = Stream(None)
        stream.buffer.feed(b'foo\n')
        def write_more():
            gruvi.sleep(0.01)
            stream.buffer.feed(b'bar\n')
            gruvi.sleep(0.01)
            stream.buffer.feed_error(RuntimeError)
        gruvi.spawn(write_more)
        it = iter(stream)
        self.assertEqual(six.next(it), b'foo\n')
        self.assertEqual(six.next(it), b'bar\n')
        self.assertRaises(RuntimeError, six.next, it)

    def test_iter_eof_and_error(self):
        stream = Stream(None)
        stream.buffer.feed(b'foo\nbar\n')
        stream.buffer.feed_eof()
        stream.buffer.feed_error(RuntimeError)
        it = iter(stream)
        self.assertEqual(six.next(it), b'foo\n')
        self.assertEqual(six.next(it), b'bar\n')
        self.assertRaises(StopIteration, six.next, it)


class TestWrappedStreamReader(UnitTest):

    def test_simple(self):
        stream = Stream(None)
        wrapped = TextIOWrapper(stream, 'utf-8')
        stream.buffer.feed(b'foo')
        self.assertEqual(wrapped.read(3), 'foo')

    def test_read_eof(self):
        stream = Stream(None)
        wrapped = TextIOWrapper(stream, 'utf-8')
        stream.buffer.feed(b'foo')
        stream.buffer.feed_eof()
        self.assertEqual(wrapped.read(), 'foo')

    def test_partial_decode_at_eof(self):
        stream = Stream(None)
        wrapped = TextIOWrapper(stream, 'utf-8')
        # \u20ac is the euro sign in case you wondered..
        buf = u'20 \u20ac'.encode('utf-8')
        stream.buffer.feed(buf[:-1])
        stream.buffer.feed_eof()
        self.assertRaises(UnicodeDecodeError, wrapped.read, 4)

    def test_partial_decode_wait(self):
        stream = Stream(None)
        wrapped = TextIOWrapper(stream, 'utf-8')
        buf = u'20 \u20ac'.encode('utf-8')
        stream.buffer.feed(buf[:-1])
        def write_last():
            gruvi.sleep(0.01)
            stream.buffer.feed(buf[-1:])
        gruvi.spawn(write_last)
        self.assertEqual(wrapped.read(4), u'20 \u20ac')


class TestStreamProtocol(UnitTest):

    def test_read(self):
        # Test that read() works.
        transport = MockTransport()
        protocol = StreamProtocol()
        transport.start(protocol)
        protocol.data_received(b'foo')
        stream = protocol.stream
        self.assertEqual(stream.read(3), b'foo')
        protocol.data_received(b'bar')
        protocol.eof_received()
        self.assertEqual(stream.read(), b'bar')

    def test_read_after_error(self):
        # Test that the buffer can be emptied after an error occurs.
        transport = MockTransport()
        protocol = StreamProtocol()
        transport.start(protocol)
        protocol.data_received(b'foobar')
        protocol.connection_lost(RuntimeError)
        stream = protocol.stream
        self.assertEqual(stream.read(3), b'foo')
        self.assertEqual(stream.read(3), b'bar')
        self.assertRaises(RuntimeError, stream.read)

    def test_readline(self):
        # Test that readline() works.
        transport = MockTransport()
        protocol = StreamProtocol()
        transport.start(protocol)
        protocol.data_received(b'foo\n')
        stream = protocol.stream
        self.assertEqual(stream.readline(), b'foo\n')

    def test_readlines(self):
        # Test that readlines() works.
        transport = MockTransport()
        protocol = StreamProtocol()
        transport.start(protocol)
        protocol.data_received(b'foo\nbar\n')
        protocol.eof_received()
        stream = protocol.stream
        self.assertEqual(stream.readlines(), [b'foo\n', b'bar\n'])

    def test_iter(self):
        # Ensure that iterating over a reader produces lines.
        transport = MockTransport()
        protocol = StreamProtocol()
        transport.start(protocol)
        protocol.data_received(b'foo\nbar\n')
        protocol.eof_received()
        it = iter(protocol.stream)
        self.assertEqual(six.next(it), b'foo\n')
        self.assertEqual(six.next(it), b'bar\n')
        self.assertRaises(StopIteration, six.next, it)

    def test_write(self):
        # Test that write() works.
        transport = MockTransport()
        protocol = StreamProtocol()
        transport.start(protocol)
        protocol.stream.write(b'foo')
        self.assertEqual(transport.buffer.getvalue(), b'foo')

    def test_writelines(self):
        # Test that writelines() works.
        transport = MockTransport()
        protocol = StreamProtocol()
        transport.start(protocol)
        protocol.stream.writelines([b'foo', b'bar'])
        self.assertEqual(transport.buffer.getvalue(), b'foobar')

    def test_write_eof(self):
        # Test that write_eof() works.
        transport = MockTransport()
        protocol = StreamProtocol()
        transport.start(protocol)
        self.assertFalse(transport.eof)
        protocol.stream.write_eof()
        self.assertTrue(transport.eof)

    def test_read_write_flow_control(self):
        # Test the read and write flow control of a stream transport.
        transport = MockTransport()
        protocol = StreamProtocol()
        transport.start(protocol)
        protocol.stream.buffer.set_buffer_limits(100)
        transport.set_write_buffer_limits(50)
        def reader():
            while True:
                buf = protocol.stream.read(20)
                if not buf:
                    break
                protocol.stream.write(buf)
        fib = gruvi.spawn(reader)
        buf = b'x' * 20
        interrupted = 0
        for i in range(100):
            protocol.data_received(buf)
            if transport._reading:
                continue
            interrupted += 1
            self.assertGreater(protocol.stream.buffer.get_buffer_size(), 0)
            # Switch to the reader() fiber which will fill up the transport
            # write buffer.
            gruvi.sleep(0)
            # The transport write buffer should be full but the protocol read
            # buffer should still contain something.
            self.assertGreater(protocol.stream.buffer.get_buffer_size(), 0)
            self.assertFalse(transport._can_write.is_set())
            # Drain write buffer and resume writing
            transport.drain()
        self.assertGreater(interrupted, 30)
        fib.cancel()
        gruvi.sleep(0)


def echo_handler(stream, transport, protocol):
    while True:
        buf = stream.readline()
        if not buf:
            break
        stream.write(buf)


class TestStreamEndpoints(UnitTest):

    def test_echo_pipe(self):
        server = StreamServer(echo_handler)
        server.listen(self.pipename())
        client = StreamClient()
        client.connect(server.addresses[0])
        client.write(b'foo\n')
        client.write_eof()
        self.assertEqual(client.readline(), b'foo\n')
        self.assertEqual(client.readline(), b'')
        server.close()
        client.close()

    def test_echo_pipe_ssl(self):
        server = StreamServer(echo_handler)
        context = self.get_ssl_context()
        server.listen(self.pipename(), ssl=context)
        client = StreamClient()
        client.connect(server.addresses[0], ssl=context)
        client.write(b'foo\n')
        self.assertEqual(client.readline(), b'foo\n')
        server.close()
        client.close()

    def test_echo_tcp(self):
        server = StreamServer(echo_handler)
        server.listen(('127.0.0.1', 0))
        client = StreamClient()
        client.connect(server.addresses[0])
        client.write(b'foo\n')
        client.write_eof()
        self.assertEqual(client.readline(), b'foo\n')
        server.close()
        client.close()

    def test_echo_tcp_ssl(self):
        server = StreamServer(echo_handler)
        context = self.get_ssl_context()
        server.listen(('127.0.0.1', 0), ssl=context)
        client = StreamClient()
        client.connect(server.addresses[0], ssl=context)
        client.write(b'foo\n')
        self.assertEqual(client.readline(), b'foo\n')
        server.close()
        client.close()

    def test_read_timeout(self):
        server = StreamServer(echo_handler)
        server.listen(self.pipename())
        client = StreamClient(timeout=0.01)
        client.connect(server.addresses[0])
        self.assertRaises(Timeout, client.readline)
        server.close()
        client.close()

    def test_echo_data(self):
        # Echo a bunch of data and ensure it is echoed identically
        server = StreamServer(echo_handler)
        server.listen(('127.0.0.1', 0))
        client = StreamClient()
        client.connect(server.addresses[0])
        md1 = hashlib.sha256()
        md2 = hashlib.sha256()
        def produce():
            for i in range(1000):
                chunk = os.urandom(1024)
                client.write(chunk)
                md1.update(chunk)
            client.write_eof()
        def consume():
            while True:
                buf = client.read1(1024)
                if not buf:
                    break
                md2.update(buf)
        f1 = gruvi.spawn(produce)
        f2 = gruvi.spawn(consume)
        f1.join(); f2.join()
        self.assertEqual(md1.digest(), md2.digest())
        server.close()
        client.close()

    def test_connection_limit(self):
        server = StreamServer(echo_handler)
        server.listen(('127.0.0.1', 0))
        addr = server.addresses[0]
        server.max_connections = 10
        clients = []
        try:
            for i in range(15):
                client = StreamClient()
                client.connect(addr)
                client.write(b'foo\n')
                buf = client.readline()
                if buf == b'':  # connection closed: EOF
                    client.close()
                clients.append(client)
        except TransportError:
            pass
        self.assertLessEqual(len(server.connections), server.max_connections)
        for client in clients:
            client.close()
        server.close()


if __name__ == '__main__':
    unittest.main()



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2014 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function, division

import time
import unittest

from gruvi.http import HttpProtocol, HttpServer, HttpClient
from support import PerformanceTest, MockTransport


def hello_app(environ, start_response):
    headers = [('Content-Type', 'text/plain')]
    start_response('200 OK', headers)
    return [b'Hello!']


class PerfHttp(PerformanceTest):

    def perf_parsing_speed(self):
        transport = MockTransport()
        protocol = HttpProtocol()
        transport.start(protocol)
        r = b'HTTP/1.1 200 OK\r\nContent-Length: 10000\r\n\r\n'
        r += b'x' * 10000
        reqs = 4 * r
        nbytes = 0
        t0 = t1 = time.time()
        while t1 - t0 < 0.2:
            protocol.data_received(reqs)
            del protocol._queue._heap[:]
            nbytes += len(reqs)
            t1 = time.time()
        speed = nbytes / (t1 - t0) / (1024 * 1024)
        self.add_result(speed)

    def perf_server_throughput(self):
        server = HttpServer(hello_app)
        server.listen(('localhost', 0))
        addr = server.addresses[0]
        client = HttpClient()
        client.connect(addr)
        nrequests = 0
        t0 = t1 = time.time()
        while t1 - t0 < 0.2:
            client.request('GET', '/')
            resp = client.getresponse()
            self.assertEqual(resp.body.read(), b'Hello!')
            nrequests += 1
            t1 = time.time()
        throughput = nrequests / (t1 - t0)
        self.add_result(throughput)
        server.close()
        client.close()


if __name__ == '__main__':
    unittest.defaultTestLoader.testMethodPrefix = 'perf'
    unittest.main()



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2014 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import sys
import time
import threading
import pyuv

from . import fibers, compat, logging, util
from .hub import switchpoint, switch_back
from .sync import Event, Queue
from .errors import Timeout, Cancelled
from .callbacks import add_callback, remove_callback, run_callbacks

__all__ = ['Future', 'PoolBase', 'FiberPool', 'ThreadPool', 'get_io_pool',
           'get_cpu_pool', 'blocking', 'wait', 'as_completed']


class Future(object):
    """The state of an asynchronous function call.

    A future captures the state and the future result of an asynchronous
    function call. Futures are not instantiated directly but are created by a
    :class:`PoolBase` implementation. The pool accepts work in the form of a
    Python function via its :meth:`~PoolBase.submit` method. This method
    returns a future representing the state and the result of the function that
    will be executed asynchronously in the pool.

    A future progresses through different states during its lifecyle:

    * Initially the future is in the pending state.
    * Once the pool has capacity to run the function, it moves to the running
      state. In this state, :meth:`~Future.running` will return ``True``.
    * Once the function completes or raises an exception, the future moves to
      the done state. In this state, :meth:`~Future.done` will return ``True``.

    The future and the asynchronous function it represents are two distinct
    entities but for brevity the terms are often used interchangeably,
    including in this manual. For example, if the asynchronous function
    completes it is said that the future has completed, and cancelling the
    future really means cancelling the asynchronous function it is capturing
    the state of.
    """

    __slots__ = ('_result', '_state', '_lock', '_done', '_callbacks')

    S_PENDING, S_RUNNING, S_RUNNING_NOCANCEL, S_DONE, S_EXCEPTION = range(5)

    def __init__(self):
        self._result = None
        self._state = self.S_PENDING
        self._lock = threading.Lock()
        self._done = Event()
        self._callbacks = None

    def running(self):
        """Return whether this future is running. """
        return self._state in (self.S_RUNNING, self.S_RUNNING_NOCANCEL)

    def cancelled(self):
        """Return whether this future was successfully cancelled."""
        return self._state == self.S_EXCEPTION and isinstance(self._result, Cancelled)

    def done(self):
        """Return whether this future is done."""
        return self._state in (self.S_DONE, self.S_EXCEPTION)

    def cancel(self):
        """Cancel the execution of the async function, if possible.

        This method marks the future as done and sets the :class:`Cancelled`
        exception.

        A future that is not running can always be cancelled. However when a
        future is running, the ability to cancel it depends on the pool
        implementation. For example, a fiber pool can cancel running fibers but
        a thread pool cannot.

        Return ``True`` if the future could be cancelled, ``False`` otherwise.
        """
        # We leverage/abuse our _done Event's thread lock as our own lock.
        # Since it's a private copy it should be OK, and it saves some memory.
        # Just be sure that we don't modify the event with the lock held.
        with self._lock:
            if self._state not in (self.S_PENDING, self.S_RUNNING):
                return False
            self._result = Cancelled('cancelled by Future.cancel()')
            self._state = self.S_EXCEPTION
            self._done.set()
            return True

    @switchpoint
    def result(self, timeout=None):
        """Wait for the future to complete and return its result.

        If the function returned normally, its return value is returned here.
        If the function raised an exception, the exception is re-raised here.
        """
        if not self._done.wait(timeout):
            raise Timeout('timeout waiting for future')
        # No more state changes after _done is set so no lock needed.
        if self._state == self.S_EXCEPTION:
            raise compat.saved_exc(self._result)
        return self._result

    @switchpoint
    def exception(self, timeout=None):
        """Wait for the async function to complete and return its exception.

        If the function did not raise an exception this returns ``None``.
        """
        if not self._done.wait(timeout):
            raise Timeout('timeout waiting for future')
        if self._state == self.S_EXCEPTION:
            return self._result

    def add_done_callback(self, callback, *args):
        """Add a callback that gets called when the future completes.

        The callback will be called in the context of the fiber that sets the
        future's result. The callback is called with the positional arguments
        *args* provided to this method.

        The return value is an opaque handle that can be used with
        :meth:`~gruvi.Future.remove_done_callback` to remove the callback.

        If the future has already completed, then the callback is called
        immediately from this method and the return value will be ``None``.
        """
        with self._lock:
            if self._state not in (self.S_DONE, self.S_EXCEPTION):
                return add_callback(self, callback, args)
        callback(*args)

    def remove_done_callback(self, handle):
        """Remove a callback that was added by :meth:`~Future.add_done_callback`.

        It is not an error to remove a callback that was already removed.
        """
        with self._lock:
            remove_callback(self, handle)

    # The following are internal functions used by the pool implementations.

    def set_running(self, can_cancel=False):
        # Set the future in the running state.
        with self._lock:
            if self._state != self.S_PENDING:
                return
            self._state = self.S_RUNNING if can_cancel else self.S_RUNNING_NOCANCEL

    def set_result(self, result):
        # Mark the future as done and set its result. If the future has already
        # completed, then this does nothing.
        with self._lock:
            if self._state in (self.S_DONE, self.S_EXCEPTION):
                return
            self._result = result
            self._state = self.S_DONE
        self._done.set()
        run_callbacks(self)

    def set_exception(self, exception):
        # Mark the future as done and set an exception. If the future has
        # already completed, then this does nothing.
        with self._lock:
            if self._state in (self.S_DONE, self.S_EXCEPTION):
                return
            self._result = exception
            self._state = self.S_EXCEPTION
        self._done.set()
        run_callbacks(self)


class PoolBase(object):
    """Base class for pools.

    A pool contains a set of workers that can execute function calls
    asynchronously.
    """

    _PoolClosing = object()

    def __init__(self, maxsize=None, minsize=0, name=None):
        """
        The *maxsize* argument specifies the maximum numbers of workers that
        will be created in the pool. If *maxsize* is ``None`` then the pool can
        grow without bound.

        Normally the pool starts with zero workers and grows up to *maxsize* on
        demand. The *minsize* parameter can be used to change this behavior an
        make sure that there will always be at least this many workers in the
        pool.

        The *name* parameter gives a name for this pool. The pool name will
        show up in log messages related to the pool.
        """
        self._maxsize = maxsize
        self._name = name or util.objref(self)
        self._minsize = minsize
        self._log = logging.get_logger()
        self._workers = set()
        # We never switch while holding the lock which means we can use a
        # threading lock which is more efficient than a fiber lock.
        self._lock = threading.Lock()
        self._queue = Queue()
        self._closing = False
        self._closed = Event()
        self._next_worker = 0
        for i in range(self._minsize):
            self._spawn_worker()
        if self._minsize:
            self._log.debug('pre-spawned {} workers', self._minsize)

    @property
    def maxsize(self):
        """The maximum size of this pool."""
        return self._maxsize

    @property
    def minsize(self):
        """The minimum size of this pool."""
        return self._minsize

    @property
    def name(self):
        """The pool name."""
        return self._name

    def _current_worker(self):
        raise NotImplementedError

    def _create_worker(self):
        raise NotImplementedError

    def _start_work(self, fut):
        raise NotImplementedError

    def _worker_main(self):
        # Main function for each worker in the pool.
        try:
            while True:
                work = None or self._queue.get()
                try:
                    if work is self._PoolClosing:
                        self._log.debug('worker shutting down')
                        break
                    func, args, fut = work
                    if fut.cancelled():
                        continue  # Future was cancelled before it could be run.
                    self._start_work(fut)
                    try:
                        result = func(*args)
                        fut.set_result(result)
                    except Cancelled as e:
                        self._log.debug('worker was cancelled ({!s})', e)
                        # The future might have cancelled itself so make sure
                        # to set the exception, possibly unnecessarily.
                        fut.set_exception(e)
                    except:
                        # OK to catch all since we will exit.
                        self._log.debug('uncaught exception in worker', exc_info=True)
                        fut.set_exception(sys.exc_info()[1])
                        break
                finally:
                    self._queue.task_done()
        finally:
            with self._lock:
                self._workers.remove(self._current_worker())
                if work is self._PoolClosing and self._workers:
                    self._queue.put_nowait(work)
                elif work is self._PoolClosing:
                    self._closed.set()
                else:
                    self._spawn_workers()

    def _spawn_worker(self):
        # Spawn a single new worker. The lock must be held.
        name = '{}:{}'.format(self._name, self._next_worker)
        self._next_worker += 1
        self._workers.add(self._create_worker(name))

    def _spawn_workers(self):
        # Spawn new workers if required. The lock must be held.
        # Note that holding the lock prevents work from being added to the
        # queue but not from being drained. The result is that the value of
        # 'tospawn' we calculate below may be too high. However, since we limit
        # the number of workers we spin up in one batch, this is not a big deal
        # and actually simplifies the overall locking significantly (we cannot
        # do a _queue._get() under a plain lock as it might block the thread).
        nworkers = len(self._workers)
        queued = self._queue.qsize()
        active_workers = self._queue.unfinished_tasks - queued
        idle_workers = nworkers - active_workers
        wanted = max(0, queued - idle_workers, self._minsize - nworkers)
        mayspawn = max(0, self._maxsize or wanted - nworkers)
        tospawn = min(10, wanted, mayspawn)
        for i in range(tospawn):
            self._spawn_worker()
        if tospawn:
            self._log.debug('spawned {} workers, total = {}', tospawn, len(self._workers))

    def submit(self, func, *args):
        """Run *func* asynchronously.

        The function is run in the pool which will run it asynchrously. The
        function is called with positional argument *args*.

        The return value is a :class:`Future` that captures the state and the
        future result of the asynchronous function call.
        """
        with self._lock:
            if self._closing:
                raise RuntimeError('pool is closing/closed')
            result = Future()
            self._queue.put_nowait((func, args, result))
            self._spawn_workers()
        return result

    @switchpoint
    def map(self, func, *iterables, **kwargs):
        """Apply *func* to the elements of the sequences in *iterables*.

        All invocations of *func* are run in the pool. If multiple iterables
        are provided, then *func* must take this many arguments, and is applied
        with one element from each iterable. All iterables must yield the same
        number of elements.

        An optional *timeout* keyword argument may be provided to specify a
        timeout.

        This returns a generator yielding the results.
        """
        with self._lock:
            if self._closing:
                raise RuntimeError('pool is closing/closed')
            timeout = kwargs.pop('timeout', None)
            futures = []
            for args in zip(*iterables):
                result = Future()
                self._queue.put_nowait((func, args, result))
                futures.append(result)
            self._spawn_workers()
        try:
            with switch_back(timeout):
                for future in futures:
                    yield future.result()
        except Exception:
            # Timeout, GeneratorExit or future.set_exception()
            for future in futures:
                if not future.done():
                    future.cancel()
            raise

    @switchpoint
    def join(self):
        """Wait until all jobs in the pool have completed.

        New submissions are not blocked. This means that if you continue adding
        work via :meth:`~PoolBase.submit` or :meth:`~PoolBase.map` then this
        method might never finish.
        """
        self._queue.join()

    @switchpoint
    def close(self):
        """Close the pool and wait for all workers to exit.

        New submissions will be blocked. Workers will exit once their current
        job is finished. This method will return after all workers have exited.
        """
        with self._lock:
            if self._closing:
                return
            self._closing = True
            if not self._workers:
                self._closed.set()
                return
            self._queue.put_nowait(self._PoolClosing)
        self._closed.wait()


class FiberPool(PoolBase):
    """A pool that uses fiber workers."""

    def __init__(self, *args, **kwargs):
        # Provided in order not to repeat the PoolBase.__init__ doc string.
        super(FiberPool, self).__init__(*args, **kwargs)

    def _current_worker(self):
        return fibers.current_fiber()

    def _create_worker(self, name):
        fiber = fibers.Fiber(self._worker_main, name=name)
        fiber.start()
        return fiber

    def _cancel_fiber(self, fut, fiber):
        if fut.cancelled():
            fiber.cancel()

    def _start_work(self, fut):
        fut.set_running(True)
        fut.add_done_callback(self._cancel_fiber, fut, self._current_worker())


Pool = FiberPool


class ThreadPool(PoolBase):
    """A pool that uses thread workers."""

    def __init__(self, *args, **kwargs):
        # Provided in order not to repeat the PoolBase.__init__ doc string.
        super(ThreadPool, self).__init__(*args, **kwargs)

    def _current_worker(self):
        return threading.current_thread()

    def _create_worker(self, name):
        thread = threading.Thread(target=self._worker_main, name=name)
        # Mark the threads in the pool as daemonic. This allows the program to
        # exit if the user did not close() the pool. This is useful especially
        # because there's implicitly created pools (the IO and thread pool) and
        # we cannot expect the user to close those.
        thread.daemon = True
        thread.start()
        return thread

    def _start_work(self, fut):
        fut.set_running(False)


# When constructing a pool it doesn't start any workers until they are needed.
# This makes it OK to instantiate the pools ahead of time.

_io_pool = ThreadPool(20, name='Io')
_cpu_pool = ThreadPool(len(pyuv.util.cpu_info()), name='Cpu')

def get_io_pool():
    """Return the thread pool for IO tasks.

    By default there is one IO thread pool per application, which is shared
    with all threads.
    """
    return _io_pool

def get_cpu_pool():
    """Return the thread pool for CPU intenstive tasks.

    By default there is one CPU thread pool per application, which it is shared
    with all threads.
    """
    return _cpu_pool


def blocking(func, *args, **kwargs):
    """Run a function that uses blocking IO.

    The function is run in the IO thread pool.
    """
    pool = get_io_pool()
    fut = pool.submit(func, *args, **kwargs)
    return fut.result()


@switchpoint
def _wait(pending, timeout):
    # An iterator/generator that waits for objects in the list *pending*,
    # yielding them as they become ready. The pending list is mutated.
    done = []
    have_items = Event()
    def callback(i):
        done.append(pending[i])
        pending[i] = None
        have_items.set()
    handles = [pending[i].add_done_callback(callback, i) for i in range(len(pending))]
    if timeout is not None:
        end_time = time.time() + timeout
    try:
        while pending:
            if timeout is not None:
                timeout = max(0, end_time - time.time())
            if not have_items.wait(timeout):
                raise Timeout('timeout waiting for objects')
            i = 0
            while i < len(done):
                yield done[i]
                i += 1
            del done[:]
            have_items.clear()
    finally:
        for i in range(len(pending)):
            if pending[i] is not None:
                pending[i].remove_done_callback(handles[i])


@switchpoint
def as_completed(objects, count=None, timeout=None):
    """Wait for one or more waitable objects, yielding them as they become
    ready.

    This is the iterator/generator version of :func:`wait`.
    """
    for obj in objects:
        if not hasattr(obj, 'add_done_callback'):
            raise TypeError('Expecting sequence of waitable objects')
    if count is None:
        count = len(objects)
    if count < 0 or count > len(objects):
        raise ValueError('count must be between 0 and len(objects)')
    if count == 0:
        return
    pending = list(objects)
    for obj in _wait(pending, timeout):
        yield obj
        count -= 1
        if count == 0:
            break


@switchpoint
def wait(objects, count=None, timeout=None):
    """Wait for one or more waitable objects.

    This method waits until *count* elements from the sequence of waitable
    objects *objects* have become ready. If *count* is ``None`` (the default),
    then wait for all objects to become ready.

    What "ready" is means depends on the object type. A waitable object is a
    objects that implements the ``add_done_callback()`` and
    ``remove_done_callback`` methods. This currently includes:

      * :class:`~gruvi.Event` - an event is ready when its internal flag is set.
      * :class:`~gruvi.Future` - a future is ready when its result is set.
      * :class:`~gruvi.Fiber` - a fiber is ready when has terminated.
      * :class:`~gruvi.Process` - a process is ready when the child has exited.
    """
    for obj in objects:
        if not hasattr(obj, 'add_done_callback'):
            raise TypeError('Expecting sequence of waitable objects')
    if count is None:
        count = len(objects)
    if count < 0 or count > len(objects):
        raise ValueError('count must be between 0 and len(objects)')
    if count == 0:
        return [], objects
    pending = list(objects)
    done = []
    try:
        for obj in _wait(pending, timeout):
            done.append(obj)
            if len(done) == count:
                break
    except Timeout:
        pass
    return done, list(filter(bool, pending))



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2014 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function, division

import time
import unittest

from gruvi.dbus import DbusClient, DbusServer
from support import PerformanceTest
from test_dbus import echo_app


class PerfDBus(PerformanceTest):

    def perf_message_throughput_pipe(self):
        # Test roundtrips of a simple method call over a Pipe
        server = DbusServer(echo_app)
        addr = 'unix:path=' + self.pipename()
        server.listen(addr)
        client = DbusClient()
        client.connect(addr)
        nmessages = 0
        t0 = t1 = time.time()
        while t1 - t0 < 0.2:
            client.call_method('bus.name', '/path', 'my.iface', 'Echo')
            t1 = time.time()
            nmessages += 1
        throughput = nmessages / (t1 - t0)
        self.add_result(throughput)
        server.close()
        client.close()

    def perf_message_throughput_tcp(self):
        # Test roundtrips of a simple method call over TCP.
        server = DbusServer(echo_app)
        addr = 'tcp:host=127.0.0.1,port=0'
        server.listen(addr)
        client = DbusClient()
        client.connect(server.addresses[0])
        nmessages = 0
        t0 = t1 = time.time()
        while t1 - t0 < 0.2:
            client.call_method('bus.name', '/path', 'my.iface', 'Echo')
            t1 = time.time()
            nmessages += 1
        throughput = nmessages / (t1 - t0)
        self.add_result(throughput)
        server.close()
        client.close()


if __name__ == '__main__':
    unittest.defaultTestLoader.testMethodPrefix = 'perf'
    unittest.main()



# Ping-pong between JSON-RPC client and server.

from gruvi.jsonrpc import JsonRpcClient, JsonRpcServer

def handler(message, transport, protocol):
    method = message.get('method')
    if method == 'ping':
        protocol.send_response(message, 'pong')

server = JsonRpcServer(handler)
server.listen(('localhost', 0))
addr = server.addresses[0]

client = JsonRpcClient()
client.connect(addr)
result = client.call_method('ping')

print('result = {}'.format(result))



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2014 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

from invoke import run, task


@task
def clean(ctx):
    run('find . -name __pycache__ | xargs rm -rf || :', echo=True)
    run('find . -name \*.so | xargs rm -f', echo=True)
    run('find . -name \*.pyc | xargs rm -f', echo=True)
    run('find . -name \*.egg-info | xargs rm -rf', echo=True)
    run('rm -rf build dist', echo=True)
    run('rm -rf docs/_build/*', echo=True)


@task(clean)
def develop(ctx):
    run('python setup.py build', echo=True)
    if develop:
        run('python setup.py develop', echo=True)


@task
def checksdist(ctx):
    from setup import version_info
    run('git ls-files | sort > files.git')
    run('rm -rf lib/*.egg-info')
    run('python setup.py sdist >/dev/null 2>&1')
    run('tar tfz dist/{name}-{version}.tar.gz'
       ' | sed -e \'s/^{name}-{version}\///\' -e \'/\/$/d\' -e \'/^$/d\''
       ' | sort > files.sdist'.format(**version_info))
    run('diff -u files.git files.sdist || true')
    run('rm files.git; rm files.sdist')


@task
def buildwheels(ctx):
    run('tox -e py27 -- python setup.py bdist_wheel')
    run('tox -e py35 -- python setup.py bdist_wheel')
    run('tox -e py36 -- python setup.py bdist_wheel')



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2017 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

from . import logging
from .sync import Queue
from .errors import Error
from .hub import get_hub
from .fibers import spawn

__all__ = ['ProtocolError', 'BaseProtocol', 'Protocol', 'DatagramProtocol',
           'MessageProtocol']


class ProtocolError(Error):
    """A protocol error."""


class BaseProtocol(object):
    """Base class for all protocols."""

    def __init__(self, timeout=None):
        """The *timeout* argument specifies a default timeout for various
        protocol operations."""
        self._timeout = timeout
        self._log = logging.get_logger()
        self._error = None

    def connection_made(self, transport):
        """Called when a connection is made."""

    def connection_lost(self, exc):
        """Called when a connection is lost."""

    def pause_writing(self):
        """Called when the write buffer in the transport has exceeded the high
        water mark. The protocol should stop writing new data."""

    def resume_writing(self):
        """Called when the write buffer in the transport has fallen below the
        low water mark. The protocol can start writing data again."""


class Protocol(BaseProtocol):
    """Base class for connection oriented protocols."""

    def data_received(self, data):
        """Called when a new chunk of data is received."""

    def eof_received(self):
        """Called when an EOF is received."""


class DatagramProtocol(BaseProtocol):
    """Base classs for datagram oriented protocols."""

    def datagram_received(self, data, addr):
        """Called when a new datagram is received."""

    def error_received(self, exc):
        """Called when an error has occurred."""


class MessageProtocol(Protocol):
    """Base class for message oriented protocols."""

    max_queue_size = 10

    def __init__(self, message_handler=None, timeout=None):
        super(MessageProtocol, self).__init__(timeout=timeout)
        self._message_handler = message_handler
        self._hub = get_hub()
        self._queue = Queue()
        self._dispatcher = None

    def _maybe_pause_transport(self):
        # Stop transport from calling data_received() if queue is at capacity
        if self._queue.qsize() >= self.max_queue_size:
            self._transport.pause_reading()

    def _maybe_resume_transport(self):
        # Resume data_received if queue has room.
        if self._queue.qsize() < self.max_queue_size:
            self._transport.resume_reading()

    def _dispatch_loop(self):
        # Call message handler for HTTP requests. Runs in a separate fiber.
        self._log.debug('dispatcher starting')
        try:
            while True:
                message = self._queue.get()
                if self._transport is None:
                    break
                self._maybe_resume_transport()
                self._message_handler(message, self._transport, self)
        finally:
            self._log.debug('dispatcher exiting, closing transport')
            if self._transport is not None:
                self._transport.close()

    def connection_made(self, transport):
        # Protocol callback
        self._transport = transport
        if self._message_handler:
            self._dispatcher = spawn(self._dispatch_loop)

    def connection_lost(self, exc):
        # Protocol callback
        if self._dispatcher:
            self._dispatcher.cancel()
        self._transport = None



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2017 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

from io import BufferedIOBase

from . import compat
from .util import delegate_method
from .sync import Event
from .errors import Error, Timeout, Cancelled
from .protocols import Protocol, ProtocolError
from .endpoints import Client, Server
from .hub import switchpoint
from .fibers import spawn

__all__ = ['StreamError', 'StreamBuffer', 'Stream', 'StreamProtocol',
           'StreamClient', 'StreamServer']


class StreamError(Error):
    """Stream error."""


class StreamBuffer(object):
    """A stream buffer.

    This is a utility class that is used to by :class:`Stream` to implement the
    read side buffering.
    """

    default_buffer_size = 65536

    def __init__(self, transport=None, timeout=None):
        self._transport = transport
        self._timeout = timeout
        self._can_read = Event()
        self._buffers = []
        self._buffer_size = 0
        self._buffer_high = self.default_buffer_size
        self._buffer_low = self.default_buffer_size // 2
        self._offset = 0
        self._eof = False
        self._error = None

    @property
    def eof(self):
        return self._buffer_size == 0 and self._eof

    @property
    def error(self):
        return self._buffer_size == 0 and self._error

    def get_buffer_size(self):
        """Return the size of the buffer."""
        return self._buffer_size

    def set_buffer_limits(self, high=None, low=None):
        """Set the low and high watermarks for the read buffer."""
        if high is None:
            high = self.default_buffer_size
        if low is None:
            low = high // 2
        self._buffer_high = high
        self._buffer_low = low

    def feed(self, data):
        """Add *data* to the buffer."""
        self._buffers.append(data)
        self._buffer_size += len(data)
        self._maybe_pause_transport()
        self._can_read.set()

    def feed_eof(self):
        """Set the EOF condition."""
        self._eof = True
        self._can_read.set()

    def feed_error(self, exc):
        """Set the error condition to *exc*."""
        self._error = exc
        self._can_read.set()

    def _maybe_resume_transport(self):
        if self._transport is None:
            return
        if self._buffer_size <= self._buffer_low:
            self._transport.resume_reading()

    def _maybe_pause_transport(self):
        if self._transport is None:
            return
        if self._buffer_size > self._buffer_high:
            self._transport.pause_reading()

    @switchpoint
    def get_chunk(self, size=-1, delim=None):
        # Get a single chunk of data. The chunk will be at most *size* bytes.
        # If *delim* is provided, then return a partial chunk if it contains
        # the delimiter.
        if size != 0 and not self._can_read.wait(self._timeout):
            raise Timeout('timeout waiting for data')
        if not self._buffers:
            return b''  # EOF or error
        # Clamp the current buffer to *size* bytes.
        endpos = len(self._buffers[0])
        if size == -1:
            size = endpos
        if self._offset + size < endpos:
            endpos = self._offset + size
        # Reduce it even further if the delimiter is found
        if delim:
            pos = self._buffers[0].find(delim, self._offset, endpos)
            if pos != -1:
                endpos = pos + len(delim)
        nbytes = endpos - self._offset
        # Try to move a buffer instead of copying.
        if self._offset == 0 and endpos == len(self._buffers[0]):
            chunk = self._buffers.pop(0)
        else:
            chunk = self._buffers[0][self._offset:endpos]
            self._offset = endpos
            if self._offset == len(self._buffers[0]):
                del self._buffers[0]
                self._offset = 0
        # Adjust buffer size and notify callback
        self._buffer_size -= nbytes
        self._maybe_resume_transport()
        # If there's no data and no error, clear the reading indicator.
        if not self._buffers and not self._eof and not self._error:
            self._can_read.clear()
        return chunk


class Stream(BufferedIOBase):
    """
    A byte stream.

    This class implements buffered, flow controlled and blocking read/write
    access on top of a transport.

    A stream works with ``bytes`` instances. To create a stream that works with
    unicode strings, you can wrap it with a :class:`io.TextIOWrapper`.
    """

    def __init__(self, transport, mode='rw', autoclose=False, timeout=None):
        """The *transport* argument specifies the transport on top of which to
        create the stream.

        The *mode* argument can be ``'r'`` for a read-only stream, ``'w'`` for
        a write-only stream, or ``'rw'`` for a read-write stream.

        The *autoclose* argument controls whether the underlying transport will
        be closed in the :meth:`close` method. Be careful with this as the
        close method is called from the :class:`io.BufferedIOBase` destructor,
        which may lead to unexpected closing of the transport when the stream
        goes out of scope.
        """
        self._transport = transport
        self._readable = 'r' in mode
        self._writable = 'w' in mode
        if self._readable:
            self._buffer = StreamBuffer(transport, timeout)
        self._autoclose = autoclose
        self._closed = False

    def readable(self):
        """Return whether the stream is readable."""
        return self._readable

    def writable(self):
        """Return whether the stream is writable."""
        return self._writable

    @property
    def closed(self):
        """Return whether the stream is closed."""
        return self._closed

    @property
    def buffer(self):
        """Return the :class:`StreamBuffer` for this stream."""
        if self._readable:
            return self._buffer

    def _check_readable(self):
        # Check that the transport is open and is readable
        if self._closed:
            raise StreamError('stream is closed')
        if not self._readable:
            raise StreamError('stream is not readable')

    def _check_writable(self):
        # Check that the transport is open and is writable
        if self._closed:
            raise StreamError('stream is closed')
        elif not self._writable:
            raise StreamError('stream is not writable')

    @switchpoint
    def read(self, size=-1):
        """Read up to *size* bytes.

        This function reads from the buffer multiple times until the requested
        number of bytes can be satisfied. This means that this function may
        block to wait for more data, even if some data is available. The only
        time a short read is returned, is on EOF or error.

        If *size* is not specified or negative, read until EOF.
        """
        self._check_readable()
        chunks = []
        bytes_read = 0
        bytes_left = size
        while True:
            chunk = self._buffer.get_chunk(bytes_left)
            if not chunk:
                break
            chunks.append(chunk)
            bytes_read += len(chunk)
            if bytes_read == size or not chunk:
                break
            if bytes_left > 0:
                bytes_left -= len(chunk)
        # If EOF was set, always return that instead of any error.
        if not chunks and not self._buffer.eof and self._buffer.error:
            raise compat.saved_exc(self._buffer.error)
        return b''.join(chunks)

    @switchpoint
    def read1(self, size=-1):
        """Read up to *size* bytes.

        This function reads from the buffer only once. It is useful in case you
        need to read a large input, and want to do so efficiently. If *size* is
        big enough, then this method will return the chunks passed into the
        memory buffer verbatim without any copying or slicing.
        """
        self._check_readable()
        chunk = self._buffer.get_chunk(size)
        if not chunk and not self._buffer.eof and self._buffer.error:
            raise compat.saved_exc(self._buffer.error)
        return chunk

    @switchpoint
    def readline(self, limit=-1, delim=b'\n'):
        """Read a single line.

        If EOF is reached before a full line can be read, a partial line is
        returned. If *limit* is specified, at most this many bytes will be read.
        """
        self._check_readable()
        chunks = []
        while True:
            chunk = self._buffer.get_chunk(limit, delim)
            if not chunk:
                break
            chunks.append(chunk)
            if chunk.endswith(delim):
                break
            if limit >= 0:
                limit -= len(chunk)
                if limit == 0:
                    break
        if not chunks and not self._buffer.eof and self._buffer.error:
            raise compat.saved_exc(self._buffer.error)
        return b''.join(chunks)

    @switchpoint
    def readlines(self, hint=-1):
        """Read lines until EOF, and return them as a list.

        If *hint* is specified, then stop reading lines as soon as the total
        size of all lines exceeds *hint*.
        """
        self._check_readable()
        lines = []
        chunks = []
        bytes_read = 0
        while True:
            chunk = self._buffer.get_chunk(-1, b'\n')
            if not chunk:
                break
            chunks.append(chunk)
            if chunk.endswith(b'\n'):
                lines.append(b''.join(chunks))
                del chunks[:]
                bytes_read += len(lines[-1])
            if hint >= 0 and bytes_read > hint:
                break
        if chunks:
            lines.append(b''.join(chunks))
        if not lines and not self._buffer.eof and self._buffer.error:
            raise compat.saved_exc(self._buffer.error)
        return lines

    @switchpoint
    def __iter__(self):
        """Generate lines until EOF is reached."""
        while True:
            line = self.readline()
            if not line:
                break
            yield line

    @switchpoint
    def write(self, data):
        """Write *data* to the transport.

        This method will block if the transport's write buffer is at capacity.
        """
        self._check_writable()
        self._transport._can_write.wait()
        self._transport.write(data)

    @switchpoint
    def writelines(self, seq):
        """Write the elements of the sequence *seq* to the transport.

        This method will block if the transport's write buffer is at capacity.
        """
        self._check_writable()
        for line in seq:
            self._transport._can_write.wait()
            self._transport.write(line)

    @switchpoint
    def write_eof(self):
        """Close the write direction of the transport.

        This method will block if the transport's write buffer is at capacity.
        """
        self._check_writable()
        self._transport._can_write.wait()
        self._transport.write_eof()

    @switchpoint
    def close(self):
        """Close the stream.

        If *autoclose* was passed to the constructor then the underlying
        transport will be closed as well.
        """
        if self._closed:
            return
        if self._autoclose:
            self._transport.close()
            self._transport._closed.wait()
        self._transport = None
        self._closed = True


class StreamProtocol(Protocol):
    """Byte stream protocol."""

    def __init__(self, timeout=None):
        """The *timeout* argument specifies a default timeout for protocol
        operations."""
        super(StreamProtocol, self).__init__(timeout=timeout)
        self._stream = None

    @property
    def stream(self):
        """A :class:`Stream` instance, providing blocking, flow controlled read
        and write access to the underlying transport."""
        return self._stream

    def connection_made(self, transport):
        # Protocol callback
        self._stream = Stream(transport, autoclose=True, timeout=self._timeout)
        self._transport = transport

    def data_received(self, data):
        # Protocol callback
        self._stream.buffer.feed(data)

    def eof_received(self):
        # Protocol callback
        self._stream.buffer.feed_eof()
        # I believe the most natural behavior for a stream is not to close the
        # write write direction when an EOF is received. Without this, a simple
        # echo server would not be possible. This behavior is similar to a
        # file. If EOF is reached on read, you can still write to it. Also is
        # is easy to implement autoclose in the stream handler if needed.
        return True  # Means do not autoclose.

    def connection_lost(self, exc=None):
        # Protocol callback.
        if exc:
            self._stream.buffer.feed_error(exc)
        else:
            # eof_received() is not guaranteed to happen. So if there is no error
            # make sure we set the EOF condition on the stream.
            self._stream.buffer.feed_eof()
        self._transport = None


class StreamClient(Client):
    """A stream client."""

    def __init__(self, timeout=None):
        super(StreamClient, self).__init__(self._create_protocol, timeout=timeout)

    def _create_protocol(self):
        return StreamProtocol(timeout=self._timeout)

    @property
    def stream(self):
        """An alias for ``self.protocol.reader``"""
        if not self.protocol:
            raise ProtocolError('not connected')
        return self.protocol.stream

    delegate_method(stream, Stream.read)
    delegate_method(stream, Stream.read1)
    delegate_method(stream, Stream.readline)
    delegate_method(stream, Stream.readlines)
    delegate_method(stream, Stream.write)
    delegate_method(stream, Stream.writelines)
    delegate_method(stream, Stream.write_eof)


class StreamServer(Server):
    """A stream server."""

    def __init__(self, stream_handler, timeout=None):
        """The *stream_handler* argument is a handler function to handle client
        connections. The handler will be called as ``stream_handler(stream,
        transport, protocol)``. The handler for each connection will run in a
        separate fiber so it can use blocking I/O on the stream. When the
        handler returns, the stream is closed.

        See :ref:`example-stream-server` for an example.
        """
        super(StreamServer, self).__init__(StreamProtocol, timeout=timeout)
        self._stream_handler = stream_handler
        self._dispatchers = {}

    def connection_made(self, transport, protocol):
        self._dispatchers[transport] = spawn(self._dispatch_stream, transport, protocol)

    def _dispatch_stream(self, transport, protocol):
        # Stream dispatcher, runs in a separate Fiber.
        self._log.debug('stream handler started')
        try:
            self._stream_handler(protocol.stream, transport, protocol)
        except Cancelled:
            self._log.debug('stream handler cancelled')
        except Exception:
            self._log.exception('uncaught exception in stream handler')
        transport.close()
        self._log.debug('stream handler exiting')

    def connection_lost(self, transport, protocol, exc=None):
        dispatcher = self._dispatchers.pop(transport)
        dispatcher.cancel()



#
# This file is part of gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2014 the gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import gc
import unittest

import gruvi
from support import UnitTest


class TestLocal(UnitTest):

    def test_basic(self):
        local = gruvi.local()
        local.foo = 'bar'
        self.assertEqual(local.foo, 'bar')
        del local.foo
        self.assertRaises(AttributeError, local.__getattr__, 'foo')
        self.assertRaises(AttributeError, local.__delattr__, 'foo')

    def test_isolation(self):
        local = gruvi.local()
        interleaved = []
        def fiber1():
            local.foo = 10
            interleaved.append(1)
            gruvi.sleep(0)
            self.assertEqual(local.foo, 10)
            local.foo = 30
            interleaved.append(1)
            gruvi.sleep(0)
            self.assertEqual(local.foo, 30)
        def fiber2():
            self.assertFalse(hasattr(local, 'foo'))
            local.foo = 20
            interleaved.append(2)
            gruvi.sleep(0)
            self.assertEqual(local.foo, 20)
            local.foo = 40
            interleaved.append(2)
            gruvi.sleep(0)
            self.assertEqual(local.foo, 40)
        f1 = gruvi.spawn(fiber1)
        f2 = gruvi.spawn(fiber2)
        f1.join(); f2.join()
        self.assertFalse(hasattr(local, 'foo'))
        self.assertEqual(interleaved, [1, 2, 1, 2])

    def test_cleanup_on_fiber_exit(self):
        local = gruvi.local()
        def fiber1():
            local.foo = 10
        f1 = gruvi.spawn(fiber1)
        f1.join()
        # Access the local object as if access was from f1
        self.assertIn('foo', local._keys[f1])
        self.assertEqual(local._keys[f1]['foo'], 10)
        del f1; gc.collect()
        self.assertEqual(len(local._keys), 0)


if __name__ == '__main__':
    unittest.main()



# Gruvi example program: a nc(1) like client, using StreamClient

import sys
import argparse
import gruvi

parser = argparse.ArgumentParser()
parser.add_argument('hostname')
parser.add_argument('port', type=int)
parser.add_argument('--ssl', action='store_true')
args = parser.parse_args()

remote = gruvi.StreamClient()
remote.connect((args.hostname, args.port), ssl=args.ssl)

stdin = gruvi.StreamClient()
stdin.connect(sys.stdin.fileno(), mode='r')
stdout = gruvi.StreamClient()
stdout.connect(sys.stdout.fileno(), mode='w')

done = gruvi.Event()

def stdin_reader():
    while True:
        buf = stdin.read1()
        if not buf:
            print('Got EOF on stdin')
            break
        remote.write(buf)
    done.set()

def remote_reader():
    while True:
        buf = remote.read1()
        if not buf:
            print('Got EOF from remote')
            break
        stdout.write(buf)
    done.set()

gruvi.spawn(stdin_reader)
gruvi.spawn(remote_reader)

done.wait()



#
# This file is part of gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2017 the gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import re
import logging
import unittest
import six

from support import UnitTest
from gruvi.logging import get_logger


class TestLogging(UnitTest):

    @classmethod
    def setUpClass(cls):
        super(TestLogging, cls).setUpClass()
        logger = logging.getLogger('gruvi_test')
        logger.setLevel(logging.DEBUG)
        logger.propagate = False
        stream = six.StringIO()
        handler = logging.StreamHandler(stream)
        template = '%(levelname)s %(message)s'
        handler.setFormatter(logging.Formatter(template))
        logger.addHandler(handler)
        cls.logger = logger
        cls.stream = stream

    @classmethod
    def get_messages(cls):
        value = cls.stream.getvalue()
        cls.stream.seek(0)
        cls.stream.truncate()
        messages = [m.rstrip() for m in value.splitlines()]
        return messages

    def test_get_logger(self):
        # Ensure that get_logger() returns the correct logger.
        logger = get_logger(name='gruvi_test')
        self.assertIs(logger._logger, self.logger)
        logger = get_logger()
        self.assertIsNot(logger._logger, self.logger)

    def test_get_logger_context(self):
        # Ensure that a different logger is returned for different contexts.
        l1 = get_logger(name='gruvi_test')
        l2 = get_logger(context='foo', name='gruvi_test')
        self.assertIsNot(l1, l2)
        l3 = get_logger(context='bar', name='gruvi_test')
        self.assertIsNot(l2, l3)

    def test_debug(self):
        # Ensure that DEBUG level output works if the level is set accordingly
        self.logger.setLevel(logging.DEBUG)
        logger = get_logger(name='gruvi_test')
        logger.debug('foo bar')
        messages = self.get_messages()
        self.assertEqual(len(messages), 1)
        self.assertIn('DEBUG', messages[0])
        self.assertIn('foo bar', messages[0])
        self.logger.setLevel(100)
        logger.debug('foo baz')
        messages = self.get_messages()
        self.assertEqual(len(messages), 0)

    def test_info(self):
        # Ensure that INFO level output works if the level is set accordingly
        self.logger.setLevel(logging.INFO)
        logger = get_logger(name='gruvi_test')
        logger.info('foo bar')
        messages = self.get_messages()
        self.assertEqual(len(messages), 1)
        self.assertIn('INFO', messages[0])
        self.assertIn('foo bar', messages[0])
        self.logger.setLevel(100)
        logger.info('foo baz')
        messages = self.get_messages()
        self.assertEqual(len(messages), 0)

    def test_warning(self):
        # Ensure that WARNING level output works if the level is set accordingly
        self.logger.setLevel(logging.WARNING)
        logger = get_logger(name='gruvi_test')
        logger.warning('foo bar')
        messages = self.get_messages()
        self.assertEqual(len(messages), 1)
        self.assertIn('WARNING', messages[0])
        self.assertIn('foo bar', messages[0])
        self.logger.setLevel(100)
        logger.warning('foo baz')
        messages = self.get_messages()
        self.assertEqual(len(messages), 0)

    def test_error(self):
        # Ensure that ERROR level output works if the level is set accordingly
        self.logger.setLevel(logging.ERROR)
        logger = get_logger(name='gruvi_test')
        logger.error('foo bar')
        messages = self.get_messages()
        self.assertEqual(len(messages), 1)
        self.assertIn('ERROR', messages[0])
        self.assertIn('foo bar', messages[0])
        self.logger.setLevel(100)
        logger.error('foo baz')
        messages = self.get_messages()
        self.assertEqual(len(messages), 0)

    def test_critical(self):
        # Ensure that CRITCAL level output works if the level is set accordingly
        self.logger.setLevel(logging.CRITICAL)
        logger = get_logger(name='gruvi_test')
        logger.critical('foo bar')
        messages = self.get_messages()
        self.assertEqual(len(messages), 1)
        self.assertIn('CRITICAL', messages[0])
        self.assertIn('foo bar', messages[0])
        self.logger.setLevel(100)
        logger.critical('foo baz')
        messages = self.get_messages()
        self.assertEqual(len(messages), 0)

    def test_exception(self):
        # Ensure that a logged exception includes a backtrace
        self.logger.setLevel(logging.ERROR)
        logger = get_logger(name='gruvi_test')
        try:
            raise ValueError('foo bar')
        except ValueError:
            logger.exception('baz qux')
        messages = self.get_messages()
        self.assertGreater(len(messages), 2)
        self.assertIn('ERROR', messages[0])
        self.assertIn('baz qux', messages[0])
        self.assertIn('Traceback', messages[1])
        self.assertIn('foo bar', messages[-1])

    re_frame = re.compile(r'\|test_[a-z]+\.py:[0-9]+\]')

    def test_debug_frame_info(self):
        # Ensure that frame info is appended when the level is DEBUG
        self.logger.setLevel(logging.DEBUG)
        logger = get_logger(name='gruvi_test')
        logger.info('foo bar')
        messages = self.get_messages()
        self.assertEqual(len(messages), 1)
        self.assertTrue(self.re_frame.search(messages[0]))
        self.logger.setLevel(logging.INFO)
        logger.info('foo bar')
        messages = self.get_messages()
        self.assertEqual(len(messages), 1)
        self.assertFalse(self.re_frame.search(messages[0]))

    def test_context(self):
        # Ensure that a logging context is added.
        self.logger.setLevel(logging.DEBUG)
        logger = get_logger('fooctx', name='gruvi_test')
        logger.debug('foo bar')
        messages = self.get_messages()
        self.assertEqual(len(messages), 1)
        self.assertIn('fooctx', messages[0])
        self.assertIn('foo bar', messages[0])
        logger = get_logger('barctx', name='gruvi_test')
        logger.debug('foo bar')
        messages = self.get_messages()
        self.assertEqual(len(messages), 1)
        self.assertNotIn('fooctx', messages[0])
        self.assertIn('barctx', messages[0])
        self.assertIn('foo bar', messages[0])


if __name__ == '__main__':
    unittest.main()



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2014 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import unittest

from gruvi import get_hub, Event
from gruvi.poll import MultiPoll, Poller, READABLE, WRITABLE
from gruvi.poll import check as check_mpoll

from support import UnitTest, socketpair


class TestMultiPoll(UnitTest):

    def test_basic(self):
        s1, s2 = socketpair()
        fd = s2.fileno()
        mp = MultiPoll(get_hub().loop, fd)
        check_mpoll(mp)
        cbargs = []
        called = Event()
        def callback(fd, events):
            cbargs.append((fd, events))
            called.set()
        mp.add_callback(READABLE, callback)
        check_mpoll(mp)
        called.wait(0.01)
        self.assertEqual(cbargs, [])
        s1.send(b'x')
        called.wait(0.1)
        self.assertEqual(cbargs, [(fd, READABLE)])
        self.assertEqual(s2.recv(10), b'x')
        del cbargs[:]; called.clear()
        called.wait(0.01)
        self.assertEqual(cbargs, [])
        mp.close()
        check_mpoll(mp)
        s1.close(); s2.close()

    def test_multiple(self):
        s1, s2 = socketpair()
        fd = s2.fileno()
        mp = MultiPoll(get_hub().loop, fd)
        cbargs = []
        called = Event()
        def callback(arg=0):
            def _callback(fd, events):
                cbargs.append((fd, events, arg))
                called.set()
            return _callback
        mp.add_callback(READABLE, callback(0))
        check_mpoll(mp)
        mp.add_callback(READABLE, callback(1))
        check_mpoll(mp)
        mp.add_callback(WRITABLE, callback(2))
        check_mpoll(mp)
        mp.add_callback(WRITABLE, callback(3))
        check_mpoll(mp)
        called.wait(0.1)
        self.assertEqual(cbargs, [(fd, WRITABLE, 2), (fd, WRITABLE, 3)])
        del cbargs[:]; called.clear()
        s1.send(b'x')
        called.wait(0.1)
        self.assertEqual(cbargs, [(fd, READABLE, 0), (fd, READABLE, 1),
                                  (fd, WRITABLE, 2), (fd, WRITABLE, 3)])
        self.assertEqual(s2.recv(10), b'x')
        mp.close()
        check_mpoll(mp)
        s1.close(); s2.close()

    def test_remove(self):
        s1, s2 = socketpair()
        fd = s2.fileno()
        mp = MultiPoll(get_hub().loop, fd)
        cbargs = []
        called = Event()
        def callback(arg=0):
            def _callback(fd, events):
                cbargs.append((fd, events, arg))
                called.set()
            return _callback
        h1 = mp.add_callback(READABLE, callback(0))
        check_mpoll(mp)
        h2 = mp.add_callback(READABLE, callback(1))
        check_mpoll(mp)
        s1.send(b'x')
        called.wait(0.1)
        self.assertEqual(cbargs, [(fd, READABLE, 0), (fd, READABLE, 1)])
        del cbargs[:]; called.clear()
        mp.remove_callback(h1)
        check_mpoll(mp)
        called.wait(0.1)
        self.assertEqual(cbargs, [(fd, READABLE, 1)])
        mp.remove_callback(h2)
        check_mpoll(mp)
        mp.close()
        check_mpoll(mp)
        s1.close(); s2.close()

    def test_update(self):
        s1, s2 = socketpair()
        fd = s2.fileno()
        mp = MultiPoll(get_hub().loop, fd)
        cbargs = []
        called = Event()
        def callback(fd, events):
            cbargs.append((fd, events))
            called.set()
        h1 = mp.add_callback(READABLE, callback)
        check_mpoll(mp)
        s1.send(b'x')
        called.wait(0.1)
        self.assertEqual(cbargs, [(fd, READABLE)])
        del cbargs[:]; called.clear()
        mp.update_callback(h1, READABLE|WRITABLE)
        check_mpoll(mp)
        s1.send(b'x')
        called.wait(0.1)
        self.assertEqual(cbargs, [(fd, READABLE|WRITABLE)])
        del cbargs[:]; called.clear()
        mp.update_callback(h1, WRITABLE)
        check_mpoll(mp)
        s1.send(b'x')
        called.wait(0.1)
        self.assertEqual(cbargs, [(fd, WRITABLE)])
        del cbargs[:]; called.clear()
        mp.update_callback(h1, 0)
        check_mpoll(mp)
        s1.send(b'x')
        called.wait(0.01)
        self.assertEqual(cbargs, [])
        mp.close()
        check_mpoll(mp)
        s1.close(); s2.close()

    def test_close(self):
        s1, s2 = socketpair()
        fd = s2.fileno()
        mp = MultiPoll(get_hub().loop, fd)
        cbargs = []
        called = Event()
        def callback(fd, events):
            cbargs.append((fd, events))
            called.set()
        h1 = mp.add_callback(READABLE, callback)
        h2 = mp.add_callback(READABLE, callback)
        s1.send(b'x')
        called.wait(0.1)
        self.assertEqual(cbargs, [(fd, READABLE), (fd, READABLE)])
        del cbargs[:]; called.clear()
        mp.close()
        called.wait(0.01)
        self.assertEqual(cbargs, [])
        self.assertRaises(RuntimeError, mp.add_callback, READABLE, callback)
        self.assertRaises(RuntimeError, mp.remove_callback, h1)
        self.assertRaises(RuntimeError, mp.remove_callback, h2)
        self.assertRaises(RuntimeError, mp.update_callback, h1, WRITABLE)
        self.assertRaises(RuntimeError, mp.update_callback, h2, WRITABLE)
        s1.close(); s2.close()


class TestPoller(UnitTest):

    def test_add_remove(self):
        poll = Poller(get_hub().loop)
        cbargs = []
        called = Event()
        def callback(fd, events):
            cbargs.append((fd, events))
            called.set()
        s1, s2 = socketpair()
        fd = s2.fileno()
        handle = poll.add_callback(fd, READABLE, callback)
        self.assertIsNotNone(handle)
        called.wait(0.01)
        self.assertEqual(cbargs, [])
        s1.send(b'x')
        called.wait(0.1)
        self.assertEqual(cbargs, [(fd, READABLE)])
        del cbargs[:]; called.clear()
        poll.remove_callback(fd, handle)
        called.wait(0.01)
        self.assertEqual(cbargs, [])
        poll.close()
        s1.close(); s2.close()

    def test_update(self):
        poll = Poller(get_hub().loop)
        cbargs = []
        called = Event()
        def callback(fd, events):
            cbargs.append((fd, events))
            called.set()
        s1, s2 = socketpair()
        fd = s2.fileno()
        handle = poll.add_callback(fd, WRITABLE, callback)
        self.assertIsNotNone(handle)
        called.wait(0.1)
        self.assertEqual(cbargs, [(fd, WRITABLE)])
        del cbargs[:]; called.clear()
        poll.update_callback(fd, handle, READABLE|WRITABLE)
        s1.send(b'x')
        called.wait(0.1)
        self.assertEqual(cbargs, [(fd, READABLE|WRITABLE)])
        del cbargs[:]; called.clear()
        poll.close()
        s1.close(); s2.close()

    def test_close(self):
        poll = Poller(get_hub().loop)
        cbargs = []
        called = Event()
        def callback(fd, events):
            cbargs.append((fd, events))
            called.set()
        s1, s2 = socketpair()
        fd = s2.fileno()
        handle = poll.add_callback(fd, READABLE, callback)
        self.assertIsNotNone(handle)
        s1.send(b'x')
        poll.close()
        called.wait(0.01)
        self.assertEqual(cbargs, [])
        self.assertRaises(RuntimeError, poll.add_callback, fd, READABLE, callback)
        self.assertRaises(RuntimeError, poll.remove_callback, fd, handle)
        self.assertRaises(RuntimeError, poll.update_callback, fd, handle, WRITABLE)
        s1.close(); s2.close()

    def test_multiple_fds(self):
        poll = Poller(get_hub().loop)
        cbargs = []
        called = Event()
        def callback(fd, events):
            cbargs.append((fd, events))
            called.set()
        s11, s12 = socketpair()
        fd1 = s12.fileno()
        poll.add_callback(fd1, READABLE, callback)
        s21, s22 = socketpair()
        fd2 = s22.fileno()
        poll.add_callback(fd2, READABLE, callback)
        s11.send(b'x')
        s21.send(b'x')
        called.wait()
        self.assertEqual(cbargs, [(fd1, READABLE), (fd2, READABLE)])
        poll.close()
        s11.close(); s12.close()
        s21.close(); s22.close()


if __name__ == '__main__':
    unittest.main()



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2017 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function


from os.path import abspath, split, join, exists

def up(path):
    return split(path)[0]

pathdir = up(up(abspath(__file__)))
vendordir = join(pathdir, 'gruvi_vendor')

if not exists(vendordir):
    # This is a development install. The setuptools "develop" command appears
    # not to correctly handle multiple entries in package_dir. Therefore, pull
    # in the vendorized packages directly from the top of the source tree.
    vendordir = join(up(pathdir), 'vendor')

__package__ = 'vendor'
__path__ = [vendordir]



# HTTP client and server example.

from gruvi.http import HttpClient, HttpServer

def handler(env, start_response):
    headers = [('Content-Type', 'text/plain; charset=UTF-8')]
    status = '200 OK'
    body = 'pong'
    start_response(status, headers)
    yield body.encode('utf-8')

server = HttpServer(handler)
server.listen(('localhost', 0))
addr = server.addresses[0]

client = HttpClient()
client.connect(addr)
client.request('GET', '/ping')

resp = client.getresponse()
assert resp.status_code == 200

body = resp.body.read()
print('result = {}'.format(body.decode(resp.charset)))



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2014 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import unittest

import gruvi
from gruvi.logging import get_logger
from gruvi import callbacks
from support import MemoryTest, sizeof


class TestMemory(MemoryTest):

    def mem_fiber(self):
        self.add_result(sizeof(gruvi.Fiber(lambda: None), exclude=('_log', '_hub', '_thread')))

    def mem_switch_back(self):
        self.add_result(sizeof(gruvi.switch_back(), exclude=('_hub', '_fiber')))

    def mem_assert_no_switchpoints(self):
        self.add_result(sizeof(gruvi.assert_no_switchpoints(), exclude=('_hub',)))

    def mem_hub(self):
        self.add_result(sizeof(gruvi.Hub(), exclude=('_log', '_thread', '_poll')))

    def mem_lock(self):
        self.add_result(sizeof(gruvi.Lock()))

    def mem_rlock(self):
        self.add_result(sizeof(gruvi.RLock()))

    def mem_event(self):
        self.add_result(sizeof(gruvi.Event()))

    def mem_condition(self):
        self.add_result(sizeof(gruvi.Condition()))

    def mem_queue(self):
        # In a Queue there's 2 conditions that share the same lock.
        queue = gruvi.Queue()
        self.add_result(sizeof(queue, exclude=('_lock',)) + sizeof(queue._lock))

    def mem_lifoqueue(self):
        self.add_result(sizeof(gruvi.LifoQueue(), exclude=('_lock',)) + sizeof(gruvi.Lock()))

    def mem_priorityqueue(self):
        self.add_result(sizeof(gruvi.PriorityQueue(), exclude=('_lock',)) + sizeof(gruvi.Lock()))

    def mem_logger(self):
        self.add_result(sizeof(get_logger(), exclude=('logger',)))

    def mem_dllist(self):
        self.add_result(sizeof(callbacks.dllist()))

    def mem_dllist_node(self):
        self.add_result(sizeof(callbacks.Node()))


if __name__ == '__main__':
    TestMemory.setup_loader()
    unittest.main()
